import glob 
import os
import os.path as osp

import imgviz
import json
import cv2
import numpy as np
import pandas as pd
from visionsuite.utils.visualizers.vis_seg import vis_seg
from athena.src.tasks.segmentation.frameworks.tensorflow.models.tf_model_v2 import TFModelv2


def get_mask_from_pred(pred, conf=0.5, contour_thres=50):
    mask = (pred > conf).astype(np.uint8)
    contours, _ = cv2.findContours(mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)
    points = []
    for contour in contours:
        if len(contour) < contour_thres:
            pass
        else:
            epsilon = 0.001 * cv2.arcLength(contour, True)
            approx = cv2.approxPolyDP(contour, epsilon, True)[:, 0, :].tolist()
            points.append(approx)

    return mask, points

patch_overlap_ratio = 0.2
patch_width = 512
patch_height = 512
dx = int((1. - patch_overlap_ratio) * patch_width)
dy = int((1. - patch_overlap_ratio) * patch_height)

compare_mask = True
weights = '/HDD/_projects/benchmark/semantic_segmentation/sungwoo_bottom/outputs/outputs/SEGMENTATION/w_patch_ratio_0.4_1_1_300/train/weights/last_weights.h5'
model = TFModelv2(model_name='deeplabv3plus', backbone='efficientnetb0', backbone_weights='imagenet', 
                  batch_size=1, width=patch_width, height=patch_height, channel=3, num_classes=4,
                  weights=weights)
classes = ['SCRATCH', 'TEAR', 'STABBED']

idx2class = {idx: cls for idx, cls in enumerate(classes)}
_classes = ['SCRATCH', 'TEAR', 'STABBED']
_idx2class = {idx: cls for idx, cls in enumerate(_classes)}
input_dir = '/HDD/_projects/benchmark/semantic_segmentation/sungwoo_bottom/datasets/split_dataset/val'
json_dir = '/HDD/_projects/benchmark/semantic_segmentation/sungwoo_bottom/datasets/split_dataset/val'
output_dir = '/HDD/_projects/benchmark/semantic_segmentation/sungwoo_bottom/tests/w_patch_ratio_0.4_1_1_300'
font_scale = 0

if not osp.exists(output_dir):
    os.makedirs(output_dir)

img_files = glob.glob(osp.join(input_dir, '*.bmp'))

results = {}
compare = {}
for img_file in img_files:
    filename = osp.split(osp.splitext(img_file)[0])[-1]
    img = cv2.imread(img_file)
    img_h, img_w, img_c = img.shape

    num_patches = 0
    idx2xyxys = {}
    for y0 in range(0, img_h, dy):
        for x0 in range(0, img_w, dx):
            num_patches += 1
            if y0 + patch_height > img_h:
                y = img_h - patch_height
            else:
                y = y0

            if x0 + patch_width > img_w:
                x = img_w - patch_width
            else:
                x = x0

            xmin, xmax, ymin, ymax = x, x + patch_width, y, y + patch_height
            preds = model(img[np.newaxis, ymin:ymax, xmin:xmax, :])[0].numpy()

            for idx in idx2class.keys():
                mask, points = get_mask_from_pred(preds[:, :, idx + 1], conf=0.6)
                
                
                for _points in points:
                    for _point in _points:
                        _point[0] += xmin
                        _point[1] += ymin
                
                    if idx not in idx2xyxys:
                        idx2xyxys[idx] = {'polygon': []}
                    
                    idx2xyxys[idx]['polygon'].append(_points)
                
            # if _classes is not None:
            #     new_idx2xyxys = {}
            #     for idx, _cls in enumerate(_classes):
            #         for jdx, cls in enumerate(classes):
            #             if cls == _cls:
            #                 new_idx2xyxys[idx] = idx2xyxys[jdx]
                
            #     idx2xyxys = new_idx2xyxys
            #     idx2class = _idx2class    
    
    color_map = imgviz.label_colormap()[1:len(idx2class) + 1]
    results.update({filename: {'idx2xyxys': idx2xyxys, 'idx2class': idx2class, 'img_file': img_file}})
    
    if compare_mask:
        _compare = vis_seg(img_file, idx2xyxys, idx2class, output_dir, color_map=color_map, json_dir=json_dir, 
                            seg_type='semantic', font_scale=font_scale,
                            compare_mask=compare_mask)
        _compare.update({"img_file": img_file})
        compare.update({filename: _compare})
    else:
        vis_seg(img_file, idx2xyxys, idx2class, output_dir, color_map, json_dir, compare_mask=compare_mask)
    
with open(osp.join(output_dir, 'preds.json'), 'w', encoding='utf-8') as json_file:
    json.dump(results, json_file, ensure_ascii=False, indent=4)

if compare_mask:
    with open(osp.join(output_dir, 'diff.json'), 'w', encoding='utf-8') as json_file:
        json.dump(compare, json_file, ensure_ascii=False, indent=4)
    
    df_compare = pd.DataFrame(compare)
        
    df_compare_pixel = df_compare.loc['diff_pixel'].T
    df_compare_pixel.fillna(0, inplace=True)
    df_compare_pixel.to_csv(osp.join(output_dir, 'diff_pixel.csv'))
    
    df_compare_pixel = df_compare.loc['diff_iou'].T
    df_compare_pixel.fillna(0, inplace=True)
    df_compare_pixel.to_csv(osp.join(output_dir, 'diff_iou.csv'))


import argparse 
import time
import os.path as osp 
import subprocess as sp
from visionsuite.cores.roboflow.ultralytics.train import train

def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--recipe-dir", required=True, type=str)
    parser.add_argument("--output-dir", required=True, type=str)

    return parser.parse_args()

if __name__ == '__main__':
    args = get_args()
    
    try:
        sp.run(['chmod', '-R', '777', args.output_dir])
        print(f"SUBPROCESS: {args.output_dir}")
    except Exception as error:
        print(f"Cannot chmod for {args.output_dir}: {error}")
        
    train(recipe_dir=args.recipe_dir)

    try:
        sp.run(['chmod', '-R', '777', args.output_dir])
        print(f"SUBPROCESS: {args.output_dir}")
    except Exception as error:
        print(f"Cannot chmod for {args.output_dir}: {error}")


# Copyright (c) ByteDance, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

"""
Mostly copy-paste from BEiT library:
https://github.com/microsoft/unilm/blob/master/beit/semantic_segmentation/mmcv_custom/layer_decay_optimizer_constructor.py
"""

import json

from mmcv.runner import OPTIMIZER_BUILDERS, DefaultOptimizerConstructor
from mmdet.utils import get_root_logger
from mmcv.runner import get_dist_info


def get_hivit_layer_id(name, num_layers, main_block):
    if name.startswith('backbone.'):
        name = name[9:]
        if name in ['cls_token', 'pos_embed', 'absolute_pos_embed']:
            return 0
        elif name.startswith('patch_embed'):
            return 0
        elif name.startswith('blocks'):
            i = int(name.split('.')[1])
            while i >= 0:
                try:
                    return main_block.index(i) + 1
                except ValueError:
                    i = i - 1
            return 0
    return num_layers - 1


@OPTIMIZER_BUILDERS.register_module()
class HiViTLayerDecayOptimizerConstructor(DefaultOptimizerConstructor):
    def _validate_cfg(self):
        pass
    
    def add_params(self, params, module, **kwargs):
        """Add all parameters of module to the params list.

        The parameters of the given module will be added to the list of param
        groups, with specific rules defined by paramwise_cfg.

        Args:
            params (list[dict]): A list of param groups, it will be modified
                in place.
            module (nn.Module): The module to be added.
        """
        logger = get_root_logger()

        parameter_groups = {}
        logger.info(f'self.paramwise_cfg is {self.paramwise_cfg}')
        if isinstance(self.paramwise_cfg, dict):
            num_layers = self.paramwise_cfg.get('num_layers') + 2
            decay_rate = self.paramwise_cfg.get('layer_decay_rate')
        else:
            num_layers, decay_rate = self.paramwise_cfg
            num_layers = num_layers + 2
        logger.info('Build HiViTLayerDecayOptimizerConstructor')

        weight_decay = self.base_wd
        main_block = [
            i for i, blk in enumerate(module.backbone.blocks) 
            if hasattr(blk, 'attn') and blk.attn is not None
        ]
        for name, param in module.named_parameters():
            if not param.requires_grad:
                continue  # frozen weights
            if len(param.shape) == 1 or name.endswith('.bias') or name in (
                    'pos_embed', 'cls_token', 'absolute_pos_embed') or 'relative_position_bias_table' in name:
                group_name = 'no_decay'
                this_weight_decay = 0.
            else:
                group_name = 'decay'
                this_weight_decay = weight_decay
            
            layer_id = get_hivit_layer_id(name, num_layers, main_block)
            group_name = f'layer_{layer_id}_{group_name}'

            if group_name not in parameter_groups:
                scale = decay_rate**(num_layers - layer_id - 1)

                parameter_groups[group_name] = {
                    'weight_decay': this_weight_decay,
                    'params': [],
                    'param_names': [],
                    'lr_scale': scale,
                    'group_name': group_name,
                    'lr': scale * self.base_lr,
                }

            parameter_groups[group_name]['params'].append(param)
            parameter_groups[group_name]['param_names'].append(name)
        rank, _ = get_dist_info()
        if rank == 0:
            to_display = {}
            for key in parameter_groups:
                to_display[key] = {
                    'param_names': parameter_groups[key]['param_names'],
                    'lr_scale': parameter_groups[key]['lr_scale'],
                    'lr': parameter_groups[key]['lr'],
                    'weight_decay': parameter_groups[key]['weight_decay'],
                }
            logger.info(f'Param groups = {json.dumps(to_display, indent=2)}')
        params.extend(parameter_groups.values())



@OPTIMIZER_BUILDERS.register_module()
class HiViT2LayerDecayOptimizerConstructor(DefaultOptimizerConstructor):
    def _validate_cfg(self):
        pass
    
    def add_params(self, params, module, **kwargs):
        """Add all parameters of module to the params list.

        The parameters of the given module will be added to the list of param
        groups, with specific rules defined by paramwise_cfg.

        Args:
            params (list[dict]): A list of param groups, it will be modified
                in place.
            module (nn.Module): The module to be added.
        """
        logger = get_root_logger()

        parameter_groups = {}
        logger.info(f'self.paramwise_cfg is {self.paramwise_cfg}')
        if isinstance(self.paramwise_cfg, dict):
            num_layers = self.paramwise_cfg.get('num_layers') + 2
            decay_rate = self.paramwise_cfg.get('layer_decay_rate')
        else:
            num_layers, decay_rate = self.paramwise_cfg
            num_layers = num_layers + 2
        logger.info('Build HiViTLayerDecayOptimizerConstructor')

        weight_decay = self.base_wd
        main_block = [
            i for i, blk in enumerate(module.backbone.blocks) 
            if hasattr(blk, 'mlp') and blk.mlp is not None
        ]
        for name, param in module.named_parameters():
            if not param.requires_grad:
                continue  # frozen weights
            if len(param.shape) == 1 or name.endswith('.bias') or name in (
                    'pos_embed', 'cls_token', 'absolute_pos_embed') or 'relative_position_bias_table' in name:
                group_name = 'no_decay'
                this_weight_decay = 0.
            else:
                group_name = 'decay'
                this_weight_decay = weight_decay
            
            layer_id = get_hivit_layer_id(name, num_layers, main_block)
            group_name = f'layer_{layer_id}_{group_name}'

            if group_name not in parameter_groups:
                scale = decay_rate**(num_layers - layer_id - 1)

                parameter_groups[group_name] = {
                    'weight_decay': this_weight_decay,
                    'params': [],
                    'param_names': [],
                    'lr_scale': scale,
                    'group_name': group_name,
                    'lr': scale * self.base_lr,
                }

            parameter_groups[group_name]['params'].append(param)
            parameter_groups[group_name]['param_names'].append(name)
        rank, _ = get_dist_info()
        if rank == 0:
            to_display = {}
            for key in parameter_groups:
                to_display[key] = {
                    'param_names': parameter_groups[key]['param_names'],
                    'lr_scale': parameter_groups[key]['lr_scale'],
                    'lr': parameter_groups[key]['lr'],
                    'weight_decay': parameter_groups[key]['weight_decay'],
                }
            logger.info(f'Param groups = {json.dumps(to_display, indent=2)}')
        params.extend(parameter_groups.values())




from __future__ import absolute_import

import numpy as np
import tensorflow as tf
import tensorflow.keras.backend as K

def _crps_tf(y_true, y_pred, factor=0.05):
    
    '''
    core of (pseudo) CRPS loss.
    
    y_true: two-dimensional arrays
    y_pred: two-dimensional arrays
    factor: importance of std term
    '''
    
    # mean absolute error
    mae = K.mean(tf.abs(y_pred - y_true))
    
    dist = tf.math.reduce_std(y_pred)
    
    return mae - factor*dist

def crps2d_tf(y_true, y_pred, factor=0.05):
    
    '''
    (Experimental)
    An approximated continuous ranked probability score (CRPS) loss function:
    
        CRPS = mean_abs_err - factor * std
        
    * Note that the "real CRPS" = mean_abs_err - mean_pairwise_abs_diff
    
     Replacing mean pairwise absolute difference by standard deviation offers
     a complexity reduction from O(N^2) to O(N*logN) 
    
    ** factor > 0.1 may yield negative loss values.
    
    Compatible with high-level Keras training methods
    
    Input
    ----------
        y_true: training target with shape=(batch_num, x, y, 1)
        y_pred: a forward pass with shape=(batch_num, x, y, 1)
        factor: relative importance of standard deviation term.
        
    '''
    
    y_pred = tf.convert_to_tensor(y_pred)
    y_true = tf.cast(y_true, y_pred.dtype)
    
    y_pred = tf.squeeze(y_pred)
    y_true = tf.squeeze(y_true)
    
    batch_num = y_pred.shape.as_list()[0]
    
    crps_out = 0
    for i in range(batch_num):
        crps_out += _crps_tf(y_true[i, ...], y_pred[i, ...], factor=factor)
        
    return crps_out/batch_num


def _crps_np(y_true, y_pred, factor=0.05):
    
    '''
    Numpy version of _crps_tf.
    '''
    
    # mean absolute error
    mae = np.nanmean(np.abs(y_pred - y_true))
    dist = np.nanstd(y_pred)
    
    return mae - factor*dist

def crps2d_np(y_true, y_pred, factor=0.05):
    
    '''
    (Experimental)
    Nunpy version of `crps2d_tf`.
    
    Documentation refers to `crps2d_tf`.
    '''
    
    y_true = np.squeeze(y_true)
    y_pred = np.squeeze(y_pred)
    
    batch_num = len(y_pred)
    
    crps_out = 0
    for i in range(batch_num):
        crps_out += _crps_np(y_true[i, ...], y_pred[i, ...], factor=factor)
        
    return crps_out/batch_num

# ========================= #
# Dice loss and variants

def dice_coef(y_true, y_pred, alpha=0., const=K.epsilon()):
    '''
    Sørensen–Dice coefficient for 2-d samples.
    
    Input
    ----------
        y_true, y_pred: predicted outputs and targets.
        const: a constant that smooths the loss gradient and reduces numerical instabilities.
        
    '''
    
    # flatten 2-d tensors
    y_true_pos = tf.reshape(y_true, [-1])
    y_pred_pos = tf.reshape(y_pred, [-1])
    
    # get true pos (TP), false neg (FN), false pos (FP).
    true_pos  = tf.reduce_sum(y_true_pos * y_pred_pos)
    false_neg = tf.reduce_sum(y_true_pos * (1-y_pred_pos))
    false_pos = tf.reduce_sum((1-y_true_pos) * y_pred_pos)
    
    # 2TP/(2TP+FP+FN) == 2TP/()
    if alpha == 0:
        coef_val = (2.0 * true_pos + const)/(2.0 * true_pos + false_pos + false_neg)
    else:
        coef_val = (2.0*true_pos + const)/(2.0*true_pos + alpha*false_neg + (1-alpha)*false_pos + const)

    
    return coef_val

def dice(y_true, y_pred, alpha=0., const=K.epsilon()):
    '''
    Sørensen–Dice Loss.
    
    dice(y_true, y_pred, const=K.epsilon())
    
    Input
    ----------
        const: a constant that smooths the loss gradient and reduces numerical instabilities.
        
    '''
    # tf tensor casting
    y_pred = tf.convert_to_tensor(y_pred)
    y_true = tf.cast(y_true, y_pred.dtype)
    
    # <--- squeeze-out length-1 dimensions.
    y_pred = tf.squeeze(y_pred)
    y_true = tf.squeeze(y_true)
    
    loss_val = 1 - dice_coef(y_true, y_pred, alpha=alpha, const=const)
    
    return loss_val

# ========================= #
# Tversky loss and variants

def tversky_coef(y_true, y_pred, alpha=0.5, const=K.epsilon()):
    '''
    Weighted Sørensen–Dice coefficient.
    
    Input
    ----------
        y_true, y_pred: predicted outputs and targets.
        const: a constant that smooths the loss gradient and reduces numerical instabilities.
        
    '''
    
    # flatten 2-d tensors
    y_true_pos = tf.reshape(y_true, [-1])
    y_pred_pos = tf.reshape(y_pred, [-1])
    
    # get true pos (TP), false neg (FN), false pos (FP).
    true_pos  = tf.reduce_sum(y_true_pos * y_pred_pos)
    false_neg = tf.reduce_sum(y_true_pos * (1-y_pred_pos))
    false_pos = tf.reduce_sum((1-y_true_pos) * y_pred_pos)
    
    # TP/(TP + a*FN + b*FP); a+b = 1
    coef_val = (true_pos + const)/(true_pos + alpha*false_neg + (1-alpha)*false_pos + const)
    
    return coef_val

def tversky(y_true, y_pred, alpha=0.5, const=K.epsilon()):
    '''
    Tversky Loss.
    
    tversky(y_true, y_pred, alpha=0.5, const=K.epsilon())
    
    ----------
    Hashemi, S.R., Salehi, S.S.M., Erdogmus, D., Prabhu, S.P., Warfield, S.K. and Gholipour, A., 2018. 
    Tversky as a loss function for highly unbalanced image segmentation using 3d fully convolutional deep networks. 
    arXiv preprint arXiv:1803.11078.
    
    Input
    ----------
        alpha: tunable parameter within [0, 1]. Alpha handles imbalance classification cases.
        const: a constant that smooths the loss gradient and reduces numerical instabilities.
        
    '''
    # tf tensor casting
    y_pred = tf.convert_to_tensor(y_pred)
    y_true = tf.cast(y_true, y_pred.dtype)
    
    # <--- squeeze-out length-1 dimensions.
    y_pred = tf.squeeze(y_pred)
    y_true = tf.squeeze(y_true)
    
    loss_val = 1 - tversky_coef(y_true, y_pred, alpha=alpha, const=const)
    
    return loss_val

def focal_tversky(y_true, y_pred, alpha=0.5, gamma=4/3, const=K.epsilon()):
    
    '''
    Focal Tversky Loss (FTL)
    
    focal_tversky(y_true, y_pred, alpha=0.5, gamma=4/3)
    
    ----------
    Abraham, N. and Khan, N.M., 2019, April. A novel focal tversky loss function with improved 
    attention u-net for lesion segmentation. In 2019 IEEE 16th International Symposium on Biomedical Imaging 
    (ISBI 2019) (pp. 683-687). IEEE.
    
    ----------
    Input
        alpha: tunable parameter within [0, 1]. Alpha handles imbalance classification cases 
        gamma: tunable parameter within [1, 3].
        const: a constant that smooths the loss gradient and reduces numerical instabilities.
        
    '''
    # tf tensor casting
    y_pred = tf.convert_to_tensor(y_pred)
    y_true = tf.cast(y_true, y_pred.dtype)
    
    # <--- squeeze-out length-1 dimensions.
    y_pred = tf.squeeze(y_pred)
    y_true = tf.squeeze(y_true)
    
    # (Tversky loss)**(1/gamma) 
    loss_val = tf.math.pow((1-tversky_coef(y_true, y_pred, alpha=alpha, const=const)), 1/gamma)
    
    return loss_val

# ========================= #
# MS-SSIM

def ms_ssim(y_true, y_pred, **kwargs):
    """
    Multiscale structural similarity (MS-SSIM) loss.
    
    ms_ssim(y_true, y_pred, **tf_ssim_kw)
    
    ----------
    Wang, Z., Simoncelli, E.P. and Bovik, A.C., 2003, November. Multiscale structural similarity for image quality assessment. 
    In The Thrity-Seventh Asilomar Conference on Signals, Systems & Computers, 2003 (Vol. 2, pp. 1398-1402). Ieee.
    
    ----------
    Input
        kwargs: keywords of `tf.image.ssim_multiscale`
                https://www.tensorflow.org/api_docs/python/tf/image/ssim_multiscale
                
        *Issues of `tf.image.ssim_multiscale`refers to:
                https://stackoverflow.com/questions/57127626/error-in-calculation-of-inbuilt-ms-ssim-function-in-tensorflow
    
    """
    
    y_pred = tf.convert_to_tensor(y_pred)
    y_true = tf.cast(y_true, y_pred.dtype)
    
    y_pred = tf.squeeze(y_pred)
    y_true = tf.squeeze(y_true)
    
    tf_ms_ssim = tf.image.ssim_multiscale(y_true, y_pred, **kwargs)
        
    return 1 - tf_ms_ssim

# ======================== #

def iou_box_coef(y_true, y_pred, mode='giou', dtype=tf.float32):
    
    """
    Inersection over Union (IoU) and generalized IoU coefficients for bounding boxes.
    
    iou_box_coef(y_true, y_pred, mode='giou', dtype=tf.float32)
    
    ----------
    Rezatofighi, H., Tsoi, N., Gwak, J., Sadeghian, A., Reid, I. and Savarese, S., 2019. 
    Generalized intersection over union: A metric and a loss for bounding box regression. 
    In Proceedings of the IEEE/CVF Conference on Computer Vision and Pattern Recognition (pp. 658-666).
    
    ----------
    Input
        y_true: the target bounding box. 
        y_pred: the predicted bounding box.
        
        Elements of a bounding box should be organized as: [y_min, x_min, y_max, x_max].

        mode: 'iou' for IoU coeff (i.e., Jaccard index);
              'giou' for generalized IoU coeff.
        
        dtype: the data type of input tensors.
               Default is tf.float32.

    """
    
    zero = tf.convert_to_tensor(0.0, dtype)
    
    # subtrack bounding box coords
    ymin_true, xmin_true, ymax_true, xmax_true = tf.unstack(y_true, 4, axis=-1)
    ymin_pred, xmin_pred, ymax_pred, xmax_pred = tf.unstack(y_pred, 4, axis=-1)
    
    # true area
    w_true = tf.maximum(zero, xmax_true - xmin_true)
    h_true = tf.maximum(zero, ymax_true - ymin_true)
    area_true = w_true * h_true
    
    # pred area
    w_pred = tf.maximum(zero, xmax_pred - xmin_pred)
    h_pred = tf.maximum(zero, ymax_pred - ymin_pred)
    area_pred = w_pred * h_pred
    
    # intersections
    intersect_ymin = tf.maximum(ymin_true, ymin_pred)
    intersect_xmin = tf.maximum(xmin_true, xmin_pred)
    intersect_ymax = tf.minimum(ymax_true, ymax_pred)
    intersect_xmax = tf.minimum(xmax_true, xmax_pred)
    
    w_intersect = tf.maximum(zero, intersect_xmax - intersect_xmin)
    h_intersect = tf.maximum(zero, intersect_ymax - intersect_ymin)
    area_intersect = w_intersect * h_intersect
    
    # IoU
    area_union = area_true + area_pred - area_intersect
    iou = tf.math.divide_no_nan(area_intersect, area_union)
    
    if mode == "iou":
        
        return iou
    
    else:
        
        # encolsed coords
        enclose_ymin = tf.minimum(ymin_true, ymin_pred)
        enclose_xmin = tf.minimum(xmin_true, xmin_pred)
        enclose_ymax = tf.maximum(ymax_true, ymax_pred)
        enclose_xmax = tf.maximum(xmax_true, xmax_pred)
        
        # enclosed area
        w_enclose = tf.maximum(zero, enclose_xmax - enclose_xmin)
        h_enclose = tf.maximum(zero, enclose_ymax - enclose_ymin)
        area_enclose = w_enclose * h_enclose
        
        # generalized IoU
        giou = iou - tf.math.divide_no_nan((area_enclose - area_union), area_enclose)

        return giou

def iou_box(y_true, y_pred, mode='giou', dtype=tf.float32):
    """
    Inersection over Union (IoU) and generalized IoU losses for bounding boxes. 
    
    iou_box(y_true, y_pred, mode='giou', dtype=tf.float32)
    
    ----------
    Rezatofighi, H., Tsoi, N., Gwak, J., Sadeghian, A., Reid, I. and Savarese, S., 2019. 
    Generalized intersection over union: A metric and a loss for bounding box regression. 
    In Proceedings of the IEEE/CVF Conference on Computer Vision and Pattern Recognition (pp. 658-666).
    
    ----------
    Input
        y_true: the target bounding box. 
        y_pred: the predicted bounding box.
        
        Elements of a bounding box should be organized as: [y_min, x_min, y_max, x_max].

        mode: 'iou' for IoU coeff (i.e., Jaccard index);
              'giou' for generalized IoU coeff.
        
        dtype: the data type of input tensors.
               Default is tf.float32.
        
    """
    
    y_pred = tf.convert_to_tensor(y_pred)
    y_pred = tf.cast(y_pred, dtype)
    
    y_true = tf.cast(y_true, dtype)
    
    y_pred = tf.squeeze(y_pred)
    y_true = tf.squeeze(y_true)

    return 1 - iou_box_coef(y_true, y_pred, mode=mode, dtype=dtype)


def iou_seg(y_true, y_pred, dtype=tf.float32):
    """
    Inersection over Union (IoU) loss for segmentation maps. 
    
    iou_seg(y_true, y_pred, dtype=tf.float32)
    
    ----------
    Rahman, M.A. and Wang, Y., 2016, December. Optimizing intersection-over-union in deep neural networks for 
    image segmentation. In International symposium on visual computing (pp. 234-244). Springer, Cham.
    
    ----------
    Input
        y_true: segmentation targets, c.f. `keras.losses.categorical_crossentropy`
        y_pred: segmentation predictions.
        
        dtype: the data type of input tensors.
               Default is tf.float32.
        
    """

    # tf tensor casting
    y_pred = tf.convert_to_tensor(y_pred)
    y_pred = tf.cast(y_pred, dtype)
    y_true = tf.cast(y_true, y_pred.dtype)

    y_pred = tf.squeeze(y_pred)
    y_true = tf.squeeze(y_true)
    
    y_true_pos = tf.reshape(y_true, [-1])
    y_pred_pos = tf.reshape(y_pred, [-1])

    area_intersect = tf.reduce_sum(tf.multiply(y_true_pos, y_pred_pos))
    
    area_true = tf.reduce_sum(y_true_pos)
    area_pred = tf.reduce_sum(y_pred_pos)
    area_union = area_true + area_pred - area_intersect
    
    return 1-tf.math.divide_no_nan(area_intersect, area_union)

# ========================= #
# Semi-hard triplet

def triplet_1d(y_true, y_pred, N, margin=5.0):
    
    '''
    (Experimental)
    Semi-hard triplet loss with one-dimensional vectors of anchor, positive, and negative.
    
    triplet_1d(y_true, y_pred, N, margin=5.0)
    
    Input
    ----------
        y_true: a dummy input, not used within this function. Appeared as a requirment of tf.keras.loss function format.
        y_pred: a single pass of triplet training, with `shape=(batch_num, 3*embeded_vector_size)`.
                i.e., `y_pred` is the ordered and concatenated anchor, positive, and negative embeddings.
        N: Size (dimensions) of embedded vectors
        margin: a positive number that prevents negative loss.
        
    '''
    
    # anchor sample pair separations.
    Embd_anchor = y_pred[:, 0:N]
    Embd_pos = y_pred[:, N:2*N]
    Embd_neg = y_pred[:, 2*N:]
    
    # squared distance measures
    d_pos = tf.reduce_sum(tf.square(Embd_anchor - Embd_pos), 1)
    d_neg = tf.reduce_sum(tf.square(Embd_anchor - Embd_neg), 1)
    loss_val = tf.maximum(0., margin + d_pos - d_neg)
    loss_val = tf.reduce_mean(loss_val)
    
    return loss_val



from __future__ import absolute_import

from keras_unet_collection.layer_utils import *
from keras_unet_collection.transformer_layers import patch_extract, patch_embedding, SwinTransformerBlock, patch_merging, patch_expanding

from tensorflow.keras.layers import Input, Dense
from tensorflow.keras.models import Model

def swin_transformer_stack(X, stack_num, embed_dim, num_patch, num_heads, window_size, num_mlp, shift_window=True, name=''):
    '''
    Stacked Swin Transformers that share the same token size.
    
    Alternated Window-MSA and Swin-MSA will be configured if `shift_window=True`, Window-MSA only otherwise.
    *Dropout is turned off.
    '''
    # Turn-off dropouts
    mlp_drop_rate = 0 # Droupout after each MLP layer
    attn_drop_rate = 0 # Dropout after Swin-Attention
    proj_drop_rate = 0 # Dropout at the end of each Swin-Attention block, i.e., after linear projections
    drop_path_rate = 0 # Drop-path within skip-connections
    
    qkv_bias = True # Convert embedded patches to query, key, and values with a learnable additive value
    qk_scale = None # None: Re-scale query based on embed dimensions per attention head # Float for user specified scaling factor
    
    if shift_window:
        shift_size = window_size // 2
    else:
        shift_size = 0
    
    for i in range(stack_num):
    
        if i % 2 == 0:
            shift_size_temp = 0
        else:
            shift_size_temp = shift_size

        X = SwinTransformerBlock(dim=embed_dim, num_patch=num_patch, num_heads=num_heads, 
                                 window_size=window_size, shift_size=shift_size_temp, num_mlp=num_mlp, qkv_bias=qkv_bias, qk_scale=qk_scale,
                                 mlp_drop=mlp_drop_rate, attn_drop=attn_drop_rate, proj_drop=proj_drop_rate, drop_path_prob=drop_path_rate, 
                                 name='name{}'.format(i))(X)
    return X


def swin_unet_2d_base(input_tensor, filter_num_begin, depth, stack_num_down, stack_num_up, 
                      patch_size, num_heads, window_size, num_mlp, shift_window=True, name='swin_unet'):
    '''
    The base of SwinUNET.
    
    ----------
    Cao, H., Wang, Y., Chen, J., Jiang, D., Zhang, X., Tian, Q. and Wang, M., 2021. 
    Swin-Unet: Unet-like Pure Transformer for Medical Image Segmentation. arXiv preprint arXiv:2105.05537.
    
    Input
    ----------
        input_tensor: the input tensor of the base, e.g., `keras.layers.Inpyt((None, None, 3))`.
        filter_num_begin: number of channels in the first downsampling block; 
                          it is also the number of embedded dimensions.
        depth: the depth of Swin-UNET, e.g., depth=4 means three down/upsampling levels and a bottom level.
        stack_num_down: number of convolutional layers per downsampling level/block. 
        stack_num_up: number of convolutional layers (after concatenation) per upsampling level/block.
        name: prefix of the created keras model and its layers.
        
        ---------- (keywords of Swin-Transformers) ----------
        
        patch_size: The size of extracted patches, 
                    e.g., patch_size=(2, 2) means 2-by-2 patches
                    *Height and width of the patch must be equal.
                    
        num_heads: number of attention heads per down/upsampling level,
                     e.g., num_heads=[4, 8, 16, 16] means increased attention heads with increasing depth.
                     *The length of num_heads must equal to `depth`.
                     
        window_size: the size of attention window per down/upsampling level,
                     e.g., window_size=[4, 2, 2, 2] means decreased window size with increasing depth.
                     
        num_mlp: number of MLP nodes.
        
        shift_window: The indicator of window shifting;
                      shift_window=True means applying Swin-MSA for every two Swin-Transformer blocks.
                      shift_window=False means MSA with fixed window locations for all blocks.

    Output
    ----------
        output tensor.
        
    Note: This function is experimental.
          The activation functions of all Swin-Transformers are fixed to GELU.
    
    '''
    # Compute number be patches to be embeded
    input_size = input_tensor.shape.as_list()[1:]
    num_patch_x = input_size[0]//patch_size[0]
    num_patch_y = input_size[1]//patch_size[1]
    
    # Number of Embedded dimensions
    embed_dim = filter_num_begin
    
    depth_ = depth
    
    X_skip = []

    X = input_tensor
    
    # Patch extraction
    X = patch_extract(patch_size)(X)

    # Embed patches to tokens
    X = patch_embedding(num_patch_x*num_patch_y, embed_dim)(X)
    
    # The first Swin Transformer stack
    X = swin_transformer_stack(X, stack_num=stack_num_down, 
                               embed_dim=embed_dim, num_patch=(num_patch_x, num_patch_y), 
                               num_heads=num_heads[0], window_size=window_size[0], num_mlp=num_mlp, 
                               shift_window=shift_window, name='{}_swin_down0'.format(name))
    X_skip.append(X)
    
    # Downsampling blocks
    for i in range(depth_-1):
        
        # Patch merging
        X = patch_merging((num_patch_x, num_patch_y), embed_dim=embed_dim, name='down{}'.format(i))(X)
        
        # update token shape info
        embed_dim = embed_dim*2
        num_patch_x = num_patch_x//2
        num_patch_y = num_patch_y//2
        
        # Swin Transformer stacks
        X = swin_transformer_stack(X, stack_num=stack_num_down, 
                                   embed_dim=embed_dim, num_patch=(num_patch_x, num_patch_y), 
                                   num_heads=num_heads[i+1], window_size=window_size[i+1], num_mlp=num_mlp, 
                                   shift_window=shift_window, name='{}_swin_down{}'.format(name, i+1))
        
        # Store tensors for concat
        X_skip.append(X)
        
    # reverse indexing encoded tensors and hyperparams
    X_skip = X_skip[::-1]
    num_heads = num_heads[::-1]
    window_size = window_size[::-1]
    
    # upsampling begins at the deepest available tensor
    X = X_skip[0]
    
    # other tensors are preserved for concatenation
    X_decode = X_skip[1:]
    
    depth_decode = len(X_decode)
    
    for i in range(depth_decode):
        
        # Patch expanding
        X = patch_expanding(num_patch=(num_patch_x, num_patch_y),
                            embed_dim=embed_dim, upsample_rate=2, return_vector=True, name='{}_swin_up{}'.format(name, i))(X)
        

        # update token shape info
        embed_dim = embed_dim//2
        num_patch_x = num_patch_x*2
        num_patch_y = num_patch_y*2
        
        # Concatenation and linear projection
        X = concatenate([X, X_decode[i]], axis=-1, name='{}_concat_{}'.format(name, i))
        X = Dense(embed_dim, use_bias=False, name='{}_concat_linear_proj_{}'.format(name, i))(X)
        
        # Swin Transformer stacks
        X = swin_transformer_stack(X, stack_num=stack_num_up, 
                           embed_dim=embed_dim, num_patch=(num_patch_x, num_patch_y), 
                           num_heads=num_heads[i], window_size=window_size[i], num_mlp=num_mlp, 
                           shift_window=shift_window, name='{}_swin_up{}'.format(name, i))
        
    # The last expanding layer; it produces full-size feature maps based on the patch size
    # !!! <--- "patch_size[0]" is used; it assumes patch_size = (size, size)
    X = patch_expanding(num_patch=(num_patch_x, num_patch_y),
                        embed_dim=embed_dim, upsample_rate=patch_size[0], return_vector=False)(X)
    
    return X


def swin_unet_2d(input_size, filter_num_begin, n_labels, depth, stack_num_down, stack_num_up, 
                      patch_size, num_heads, window_size, num_mlp, output_activation='Softmax', shift_window=True, name='swin_unet'):
    '''
    The base of SwinUNET.
    
    ----------
    Cao, H., Wang, Y., Chen, J., Jiang, D., Zhang, X., Tian, Q. and Wang, M., 2021. 
    Swin-Unet: Unet-like Pure Transformer for Medical Image Segmentation. arXiv preprint arXiv:2105.05537.
    
    Input
    ----------
        input_size: the size/shape of network input, e.g., `(128, 128, 3)`.
        filter_num_begin: number of channels in the first downsampling block; 
                          it is also the number of embedded dimensions.
        n_labels: number of output labels.
        depth: the depth of Swin-UNET, e.g., depth=4 means three down/upsampling levels and a bottom level.
        stack_num_down: number of convolutional layers per downsampling level/block. 
        stack_num_up: number of convolutional layers (after concatenation) per upsampling level/block.
        name: prefix of the created keras model and its layers.
        
        ---------- (keywords of Swin-Transformers) ----------
        
        patch_size: The size of extracted patches, 
                    e.g., patch_size=(2, 2) means 2-by-2 patches
                    *Height and width of the patch must be equal.
                    
        num_heads: number of attention heads per down/upsampling level,
                     e.g., num_heads=[4, 8, 16, 16] means increased attention heads with increasing depth.
                     *The length of num_heads must equal to `depth`.
                     
        window_size: the size of attention window per down/upsampling level,
                     e.g., window_size=[4, 2, 2, 2] means decreased window size with increasing depth.
                     
        num_mlp: number of MLP nodes.
        
        shift_window: The indicator of window shifting;
                      shift_window=True means applying Swin-MSA for every two Swin-Transformer blocks.
                      shift_window=False means MSA with fixed window locations for all blocks.
        
    Output
    ----------
        model: a keras model.
    
    Note: This function is experimental.
          The activation functions of all Swin-Transformers are fixed to GELU.
    '''
    IN = Input(input_size)
    
    # base    
    X = swin_unet_2d_base(IN, filter_num_begin=filter_num_begin, depth=depth, stack_num_down=stack_num_down, stack_num_up=stack_num_up, 
                          patch_size=patch_size, num_heads=num_heads, window_size=window_size, num_mlp=num_mlp, shift_window=shift_window, name=name)
    
    # output layer
    OUT = CONV_output(X, n_labels, kernel_size=1, activation=output_activation, name='{}_output'.format(name))
    
    # functional API model
    model = Model(inputs=[IN,], outputs=[OUT,], name='{}_model'.format(name))
    
    return model



from .root import (RUNNERS, MODELS, OPTIMIZERS, DATASETS, 
                   DATALOADERS, LOSSES, SCHEDULERS, LOOPS, 
                   PIPELINES, SAMPLERS, FUNCTIONALS,
                   TRAINERS, VALIDATORS, TESTERS)
from .registry import Registry

__all__ = ['Registry', 'RUNNERS', 'MODELS', 'OPTIMIZERS', 'DATASETS', 
           'DATALOADERS', 'LOSSES', 'SCHEDULERS', 'LOOPS', 'PIPELINES',
           'SAMPLERS', 'FUNCTIONALS', 'TRAINERS', 'VALIDATORS', 'TESTERS']


import copy
import time
from contextlib import contextmanager
from typing import Generator, Optional

from .manager import ManagerMixin, _accquire_lock, _release_lock


class DefaultScope(ManagerMixin):
    """
    Scope of current task used to reset the current registry, which can be accessed globally.

    Consider the case of resetting the current ``Registry`` by
    ``default_scope`` in the internal module which cannot access runner
    directly, it is difficult to get the ``default_scope`` defined in
    ``Runner``. However, if ``Runner`` created ``DefaultScope`` instance
    by given ``default_scope``, the internal module can get
    ``default_scope`` by ``DefaultScope.get_current_instance`` everywhere.

    Args:
        name (str): Name of default scope for global access.
        scope_name (str): Scope of current task.

    Examples:
        >>> from mmengine.model import MODELS
        >>> # Define default scope in runner.
        >>> DefaultScope.get_instance('task', scope_name='mmdet')
        >>> # Get default scope globally.
        >>> scope_name = DefaultScope.get_instance('task').scope_name
    """

    def __init__(self, name: str, scope_name: str):
        super().__init__(name)
        assert isinstance(scope_name, str), (f'scope_name should be a string, but got {scope_name}')
        self._scope_name = scope_name

    @property
    def scope_name(self) -> str:
        """
        Returns:
            str: Get current scope.
        """
        return self._scope_name

    @classmethod
    def get_current_instance(cls) -> Optional['DefaultScope']:
        """
        Get latest created default scope.

        Since default_scope is an optional argument for ``Registry.build``.
        ``get_current_instance`` should return ``None`` if there is no
        ``DefaultScope`` created.

        Examples:
            >>> default_scope = DefaultScope.get_current_instance()
            >>> # There is no `DefaultScope` created yet,
            >>> # `get_current_instance` return `None`.
            >>> default_scope = DefaultScope.get_instance(
            >>>     'instance_name', scope_name='mmengine')
            >>> default_scope.scope_name
            mmengine
            >>> default_scope = DefaultScope.get_current_instance()
            >>> default_scope.scope_name
            mmengine

        Returns:
            Optional[DefaultScope]: Return None If there has not been
            ``DefaultScope`` instance created yet, otherwise return the
            latest created DefaultScope instance.
        """
        _accquire_lock()
        if cls._instance_dict:
            instance = super().get_current_instance()
        else:
            instance = None
        _release_lock()
        return instance

    @classmethod
    @contextmanager
    def overwrite_default_scope(cls, scope_name: Optional[str]) -> Generator:
        """Overwrite the current default scope with `scope_name`"""
        if scope_name is None:
            yield
        else:
            tmp = copy.deepcopy(cls._instance_dict)
            # To avoid create an instance with the same name.
            time.sleep(1e-6)
            cls.get_instance(f'overwrite-{time.time()}', scope_name=scope_name)
            try:
                yield
            finally:
                cls._instance_dict = tmp



########################################################
#                                                      #
#       author: omitted for anonymous submission       #
#                                                      #
#     credits and copyright coming upon publication    #
#                                                      #
########################################################


from torch.utils.data import DataLoader

from src import preprocessing
from src.datasets import Cityscapes


def prepare_data(args, ckpt_dir=None, with_input_orig=False, split=None):
    train_preprocessor_kwargs = {}
    if args.dataset == "cityscapes":
        Dataset = Cityscapes
        dataset_kwargs = {"n_classes": 19}
        valid_set = "val"
    else:
        raise ValueError(f"Unknown dataset: `{args.dataset}`")
    if args.aug_scale_min != 1 or args.aug_scale_max != 1.4:
        train_preprocessor_kwargs["train_random_rescale"] = (
            args.aug_scale_min,
            args.aug_scale_max,
        )

    if split in ["valid", "val", "test"]:
        valid_set = split

    # train data
    train_data = Dataset(
        data_dir=args.dataset_dir,
        split="train",
        with_input_orig=with_input_orig,
        overfit=args.overfit,
        classes=args.num_classes,
        **dataset_kwargs,
    )

    valid_data = Dataset(
        data_dir=args.dataset_dir,
        split=valid_set,
        with_input_orig=with_input_orig,
        overfit=args.overfit,
        classes=args.num_classes,
        **dataset_kwargs,
    )

    test_data = Dataset(
        data_dir=args.dataset_dir,
        split="test",
        with_input_orig=with_input_orig,
        overfit=args.overfit,
        classes=args.num_classes,
        **dataset_kwargs,
    )

    train_preprocessor = preprocessing.get_preprocessor(
        height=args.height,
        width=args.width,
        phase="train",
        **train_preprocessor_kwargs,
    )

    train_data.preprocessor = train_preprocessor

    # valid data
    valid_preprocessor = preprocessing.get_preprocessor(
        height=args.height,
        width=args.width,
        phase="test",
    )

    if args.valid_full_res:
        valid_preprocessor_full_res = preprocessing.get_preprocessor(
            phase="test",
        )

    valid_data.preprocessor = valid_preprocessor
    test_data.preprocessor = valid_preprocessor

    if args.dataset_dir is None:
        # no path to the actual data was passed -> we cannot create dataloader,
        # return the valid dataset and preprocessor object for inference only
        if args.valid_full_res:
            return valid_data, valid_preprocessor_full_res
        else:
            return valid_data, valid_preprocessor

    if args.overfit:
        args.batch_size = 2
        args.batch_size_valid = 2

        # create the data loaders
        train_loader = DataLoader(
            train_data,
            batch_size=args.batch_size,
            num_workers=args.workers,
            drop_last=True,
            shuffle=False,
        )

    # create the data loaders
    train_loader = DataLoader(
        train_data,
        batch_size=args.batch_size,
        num_workers=args.workers,
        drop_last=True,
        shuffle=True,
    )

    # for validation we can use higher batch size as activations do not
    # need to be saved for the backwards pass
    batch_size_valid = args.batch_size_valid or args.batch_size
    valid_loader = DataLoader(
        valid_data, batch_size=batch_size_valid, num_workers=args.workers, shuffle=False
    )

    test_loader = DataLoader(
        test_data, batch_size=batch_size_valid, num_workers=args.workers, shuffle=False
    )

    # count_classes(train_loader, valid_loader)

    return train_loader, valid_loader, test_loader


import torch


def count_classes(train_loader, valid_loader):
    train_count = torch.zeros(20)
    val_count = torch.zeros(20)
    for data in iter(train_loader):
        label = torch.unique(data["label"]).long()
        train_count[label] += 1
    for data in iter(valid_loader):
        label = torch.unique(data["label"]).long()
        val_count[label] += 1
    import ipdb;ipdb.set_trace()  # fmt: skip



#!/usr/bin/env python3

"""Extend and add tranformations.

Skipped:
- RandomCutOut
- RandomMosaic
"""

import numpy as np

import mmcv
from mmseg.datasets.pipelines.transforms import (
    Resize as MMSEG_Resize,
    Pad as MMSEG_Pad,
    RandomRotate as MMSEG_RandomRotate,
    RandomCrop as MMSEG_RandomCrop,
)

from ..builder import PIPELINES


@PIPELINES.register_module()
class AddIgnoreBorder(object):
    """Add ignore border to the semantic segmentation map.

    This is only useful when the dataset is not preprocessed to ignore
    the borders.

    Works for both segmentation and boundary maps.
    """

    def __init__(self, width=10, ignore_label=255):
        self.width = width
        self.label = ignore_label

    def __call__(self, results):
        for key in results.get("seg_fields", []):
            # TODO: should warn if label is in the segmentation map
            gt_seg = results[key]
            gt_seg[0 : self.width, :] = self.label
            gt_seg[-self.width :, :] = self.label
            gt_seg[:, 0 : self.width] = self.label
            gt_seg[:, -self.width :] = self.label
            results[key] = gt_seg
        return results

    def __repr__(self):
        return self.__class__.__name__ + f"(width={self.width})"


@PIPELINES.register_module(force=True)
class Resize(MMSEG_Resize):

    _supported_types = ("gt_semantic_seg", "gt_semantic_edge", "gt_inst_seg")

    def _resize_seg(self, results):
        for key in results.get("seg_fields", []):

            assert (
                key in self._supported_types
            ), f"ERR: {key} is not in {self._supported_types}"

            """Semantic edge:
            - should work fine for most cases
            - however, the edges might be thicker than how it was preprocessed
            """

            """Instance segmentation:
            - formatting matters for cv2 input
            - https://stackoverflow.com/questions/15245262/opencv-mat-element-types-and-their-sizes
            - 32-bit signed integer should work (CV_32S)
            """

            if self.keep_ratio:
                gt_seg = mmcv.imrescale(
                    results[key], results["scale"], interpolation="nearest"
                )
            else:
                gt_seg = mmcv.imresize(
                    results[key], results["scale"], interpolation="nearest"
                )
            results[key] = gt_seg


@PIPELINES.register_module(force=True)
class Pad(MMSEG_Pad):
    def __init__(
        self,
        size=None,
        size_divisor=None,
        pad_val=0,
        seg_pad_val=255,  # NOTE: 255 == background / ignore
        edge_pad_val=0,  # NOTE: 0 == background
        inst_pad_val=0,  # NOTE: 0 == background
    ):
        super().__init__(
            size=size,
            size_divisor=size_divisor,
            pad_val=pad_val,
            seg_pad_val=seg_pad_val,
        )
        self.edge_pad_val = edge_pad_val
        self.inst_pad_val = inst_pad_val

    def _pad_seg(self, results):
        for key in results.get("seg_fields", []):
            if key == "gt_semantic_edge":
                results[key] = mmcv.impad(
                    results[key],
                    shape=results["pad_shape"][:2],
                    pad_val=self.edge_pad_val,
                )
            elif key == "gt_inst_seg":
                results[key] = mmcv.impad(
                    results[key],
                    shape=results["pad_shape"][:2],
                    pad_val=self.inst_pad_val,
                )
            elif key == "gt_semantic_seg":
                results[key] = mmcv.impad(
                    results[key],
                    shape=results["pad_shape"][:2],
                    pad_val=self.seg_pad_val,
                )
            else:
                raise ValueError(f"{key} is not supported")


@PIPELINES.register_module(force=True)
class RandomRotate(MMSEG_RandomRotate):
    def __init__(
        self,
        prob,
        degree,
        pad_val=0,
        seg_pad_val=255,  # NOTE: 255 == background / ignore
        center=None,
        auto_bound=False,
        edge_pad_val=0,  # NOTE: 0 == background
        inst_pad_val=0,  # NOTE: 0 == background
    ):
        super().__init__(
            prob=prob,
            degree=degree,
            pad_val=pad_val,
            seg_pad_val=seg_pad_val,
            center=center,
            auto_bound=auto_bound,
        )
        self.edge_pad_val = (edge_pad_val,)
        self.inst_pad_val = inst_pad_val

    def __call__(self, results):
        rotate = True if np.random.rand() < self.prob else False
        degree = np.random.uniform(min(*self.degree), max(*self.degree))
        if rotate:
            # rotate image
            results["img"] = mmcv.imrotate(
                results["img"],
                angle=degree,
                border_value=self.pal_val,
                center=self.center,
                auto_bound=self.auto_bound,
            )

            # rotate segs
            for key in results.get("seg_fields", []):
                if key == "gt_semantic_edge":
                    results[key] = mmcv.imrotate(
                        results[key],
                        angle=degree,
                        border_value=self.edge_pad_val,
                        center=self.center,
                        auto_bound=self.auto_bound,
                        interpolation="nearest",
                    )
                elif key == "gt_inst_seg":
                    results[key] = mmcv.imrotate(
                        results[key],
                        angle=degree,
                        border_value=self.inst_pad_val,
                        center=self.center,
                        auto_bound=self.auto_bound,
                        interpolation="nearest",
                    )
                elif key == "gt_semantic_seg":
                    results[key] = mmcv.imrotate(
                        results[key],
                        angle=degree,
                        border_value=self.seg_pad_val,
                        center=self.center,
                        auto_bound=self.auto_bound,
                        interpolation="nearest",
                    )
                else:
                    raise ValueError(f"ERR: {key} is not supported")
        return results


@PIPELINES.register_module(force=True)
class RandomCrop(MMSEG_RandomCrop):
    """Random crop the image & seg & edge.
    Args:
        crop_size (tuple): Expected size after cropping, (h, w).
        cat_max_ratio (float): The maximum ratio that single category could
            occupy.
    """

    def __init__(self, crop_size, cat_max_ratio=1.0, ignore_index=255):
        assert crop_size[0] > 0 and crop_size[1] > 0
        self.crop_size = crop_size
        self.cat_max_ratio = cat_max_ratio
        self.ignore_index = ignore_index

    def get_crop_bbox(self, img):
        """Randomly get a crop bounding box."""
        margin_h = max(img.shape[0] - self.crop_size[0], 0)
        margin_w = max(img.shape[1] - self.crop_size[1], 0)
        offset_h = np.random.randint(0, margin_h + 1)
        offset_w = np.random.randint(0, margin_w + 1)
        crop_y1, crop_y2 = offset_h, offset_h + self.crop_size[0]
        crop_x1, crop_x2 = offset_w, offset_w + self.crop_size[1]

        return crop_y1, crop_y2, crop_x1, crop_x2

    def crop(self, img, crop_bbox):
        """Crop from ``img``"""
        crop_y1, crop_y2, crop_x1, crop_x2 = crop_bbox
        img = img[crop_y1:crop_y2, crop_x1:crop_x2, ...]
        return img

    def __call__(self, results):
        """Call function to randomly crop images, semantic segmentation maps.
        Args:
            results (dict): Result dict from loading pipeline.
        Returns:
            dict: Randomly cropped results, 'img_shape' key in result dict is
                updated according to crop size.
        """

        img = results["img"]
        crop_bbox = self.get_crop_bbox(img)

        if self.cat_max_ratio < 1.0:
            if results.get("gt_semantic_seg", None) is not None:
                # Repeat 10 times
                for _ in range(10):
                    seg_temp = self.crop(results["gt_semantic_seg"], crop_bbox)
                    labels, cnt = np.unique(seg_temp, return_counts=True)
                    cnt = cnt[labels != self.ignore_index]
                    if len(cnt) > 1 and np.max(cnt) / np.sum(cnt) < self.cat_max_ratio:
                        break
                    crop_bbox = self.get_crop_bbox(img)
            else:
                # fall back to edge
                for _ in range(10):
                    edge_temp = self.crop(results["gt_semantic_edge"], crop_bbox)
                    labels, cnt = np.unique(edge_temp, return_counts=True)
                    if len(cnt) > 1 and np.max(cnt) / np.sum(cnt) < self.cat_max_ratio:
                        break
                    crop_bbox = self.get_crop_bbox(img)

        # crop the image
        img = self.crop(img, crop_bbox)
        img_shape = img.shape
        results["img"] = img
        results["img_shape"] = img_shape

        # crop semantic seg
        for key in results.get("seg_fields", []):
            results[key] = self.crop(results[key], crop_bbox)

        return results

    def __repr__(self):
        return self.__class__.__name__ + f"(crop_size={self.crop_size})"



# Copyright (c) OpenMMLab. All rights reserved.
import glob
import os
import os.path as osp
import re
import tempfile
import time
import warnings
import zipfile
from collections import defaultdict
from functools import partial

import mmcv
import numpy as np
import torch
from mmcv.ops import nms_rotated
from mmdet.datasets.custom import CustomDataset

from mmrotate.core import eval_rbbox_map, obb2poly_np, poly2obb_np
from mmrotate.datasets.builder import ROTATED_DATASETS

@ROTATED_DATASETS.register_module()
class CustomDOTADataset(CustomDataset):
    """DOTA dataset for detection.

    Args:
        ann_file (str): Annotation file path.
        pipeline (list[dict]): Processing pipeline.
        version (str, optional): Angle representations. Defaults to 'oc'.
        difficulty (bool, optional): The difficulty threshold of GT.
    """
    CLASSES = ('plane', 'baseball-diamond', 'bridge', 'ground-track-field',
               'small-vehicle', 'large-vehicle', 'ship', 'tennis-court',
               'basketball-court', 'storage-tank', 'soccer-ball-field',
               'roundabout', 'harbor', 'swimming-pool', 'helicopter')

    PALETTE = [(165, 42, 42), (189, 183, 107), (0, 255, 0), (255, 0, 0),
               (138, 43, 226), (255, 128, 0), (255, 0, 255), (0, 255, 255),
               (255, 193, 193), (0, 51, 153), (255, 250, 205), (0, 139, 139),
               (255, 255, 0), (147, 116, 116), (0, 0, 255)]

    def __init__(self,
                 ann_file,
                 pipeline,
                 version='oc',
                 difficulty=100,
                 **kwargs):
        self.version = version
        self.difficulty = difficulty

        super(CustomDOTADataset, self).__init__(ann_file, pipeline, **kwargs)

    def __len__(self):
        """Total number of samples of data."""
        return len(self.data_infos)

    def load_annotations(self, ann_folder):
        """
            Args:
                ann_folder: folder that contains DOTA v1 annotations txt files
        """
        cls_map = {c: i
                   for i, c in enumerate(self.CLASSES)
                   }  # in mmdet v2.0 label is 0-based
        ann_files = glob.glob(ann_folder + '/*.txt')
        data_infos = []
        if not ann_files:  # test phase
            ann_files = glob.glob(ann_folder + '/*.bmp')
            for ann_file in ann_files:
                data_info = {}
                img_id = osp.split(ann_file)[1][:-4]
                img_name = img_id + '.bmp'
                data_info['filename'] = img_name
                data_info['ann'] = {}
                data_info['ann']['bboxes'] = []
                data_info['ann']['labels'] = []
                data_infos.append(data_info)
        else:
            for ann_file in ann_files:
                data_info = {}
                img_id = osp.split(ann_file)[1][:-4]
                img_name = img_id + '.bmp'
                data_info['filename'] = img_name
                data_info['ann'] = {}
                gt_bboxes = []
                gt_labels = []
                gt_polygons = []
                gt_bboxes_ignore = []
                gt_labels_ignore = []
                gt_polygons_ignore = []

                if os.path.getsize(ann_file) == 0 and self.filter_empty_gt:
                    continue

                with open(ann_file) as f:
                    s = f.readlines()
                    for si in s:
                        bbox_info = si.split()
                        poly = np.array(bbox_info[:8], dtype=np.float32)
                        try:
                            x, y, w, h, a = poly2obb_np(poly, self.version)
                        except:  # noqa: E722
                            continue
                        cls_name = bbox_info[8]
                        difficulty = int(bbox_info[9])
                        label = cls_map[cls_name]
                        if difficulty > self.difficulty:
                            pass
                        else:
                            gt_bboxes.append([x, y, w, h, a])
                            gt_labels.append(label)
                            gt_polygons.append(poly)

                if gt_bboxes:
                    data_info['ann']['bboxes'] = np.array(
                        gt_bboxes, dtype=np.float32)
                    data_info['ann']['labels'] = np.array(
                        gt_labels, dtype=np.int64)
                    data_info['ann']['polygons'] = np.array(
                        gt_polygons, dtype=np.float32)
                else:
                    data_info['ann']['bboxes'] = np.zeros((0, 5),
                                                          dtype=np.float32)
                    data_info['ann']['labels'] = np.array([], dtype=np.int64)
                    data_info['ann']['polygons'] = np.zeros((0, 8),
                                                            dtype=np.float32)

                if gt_polygons_ignore:
                    data_info['ann']['bboxes_ignore'] = np.array(
                        gt_bboxes_ignore, dtype=np.float32)
                    data_info['ann']['labels_ignore'] = np.array(
                        gt_labels_ignore, dtype=np.int64)
                    data_info['ann']['polygons_ignore'] = np.array(
                        gt_polygons_ignore, dtype=np.float32)
                else:
                    data_info['ann']['bboxes_ignore'] = np.zeros(
                        (0, 5), dtype=np.float32)
                    data_info['ann']['labels_ignore'] = np.array(
                        [], dtype=np.int64)
                    data_info['ann']['polygons_ignore'] = np.zeros(
                        (0, 8), dtype=np.float32)

                data_infos.append(data_info)

        self.img_ids = [*map(lambda x: x['filename'][:-4], data_infos)]
        return data_infos

    def _filter_imgs(self):
        """Filter images without ground truths."""
        valid_inds = []
        for i, data_info in enumerate(self.data_infos):
            if (not self.filter_empty_gt
                    or data_info['ann']['labels'].size > 0):
                valid_inds.append(i)
        return valid_inds

    def _set_group_flag(self):
        """Set flag according to image aspect ratio.

        All set to 0.
        """
        self.flag = np.zeros(len(self), dtype=np.uint8)

    def evaluate(self,
                 results,
                 metric='mAP',
                 logger=None,
                 proposal_nums=(100, 300, 1000),
                 iou_thr=0.5,
                 scale_ranges=None,
                 nproc=4):
        """Evaluate the dataset.

        Args:
            results (list): Testing results of the dataset.
            metric (str | list[str]): Metrics to be evaluated.
            logger (logging.Logger | None | str): Logger used for printing
                related information during evaluation. Default: None.
            proposal_nums (Sequence[int]): Proposal number used for evaluating
                recalls, such as recall@100, recall@1000.
                Default: (100, 300, 1000).
            iou_thr (float | list[float]): IoU threshold. It must be a float
                when evaluating mAP, and can be a list when evaluating recall.
                Default: 0.5.
            scale_ranges (list[tuple] | None): Scale ranges for evaluating mAP.
                Default: None.
            nproc (int): Processes used for computing TP and FP.
                Default: 4.
        """
        nproc = min(nproc, os.cpu_count())
        if not isinstance(metric, str):
            assert len(metric) == 1
            metric = metric[0]
        allowed_metrics = ['mAP']
        if metric not in allowed_metrics:
            raise KeyError(f'metric {metric} is not supported')
        annotations = [self.get_ann_info(i) for i in range(len(self))]
        eval_results = {}
        if metric == 'mAP':
            assert isinstance(iou_thr, float)
            mean_ap, _ = eval_rbbox_map(
                results,
                annotations,
                scale_ranges=scale_ranges,
                iou_thr=iou_thr,
                dataset=self.CLASSES,
                logger=logger,
                nproc=nproc)
            eval_results['mAP'] = mean_ap
        else:
            raise NotImplementedError

        return eval_results

    def merge_det(self, results, nproc=4):
        """Merging patch bboxes into full image.

        Args:
            results (list): Testing results of the dataset.
            nproc (int): number of process. Default: 4.

        Returns:
            list: merged results.
        """

        def extract_xy(img_id):
            """Extract x and y coordinates from image ID.

            Args:
                img_id (str): ID of the image.

            Returns:
                Tuple of two integers, the x and y coordinates.
            """
            pattern = re.compile(r'__(\d+)___(\d+)')
            match = pattern.search(img_id)
            if match:
                x, y = int(match.group(1)), int(match.group(2))
                return x, y
            else:
                warnings.warn(
                    "Can't find coordinates in filename, "
                    'the coordinates will be set to (0,0) by default.',
                    category=Warning)
                return 0, 0

        collector = defaultdict(list)
        for idx, img_id in enumerate(self.img_ids):
            result = results[idx]
            oriname = img_id.split('__', maxsplit=1)[0]
            x, y = extract_xy(img_id)
            new_result = []
            for i, dets in enumerate(result):
                bboxes, scores = dets[:, :-1], dets[:, [-1]]
                ori_bboxes = bboxes.copy()
                ori_bboxes[..., :2] = ori_bboxes[..., :2] + np.array(
                    [x, y], dtype=np.float32)
                labels = np.zeros((bboxes.shape[0], 1)) + i
                new_result.append(
                    np.concatenate([labels, ori_bboxes, scores], axis=1))
            new_result = np.concatenate(new_result, axis=0)
            collector[oriname].append(new_result)

        merge_func = partial(_merge_func, CLASSES=self.CLASSES, iou_thr=0.1)
        if nproc <= 1:
            print('Executing on Single Processor')
            merged_results = mmcv.track_iter_progress(
                (map(merge_func, collector.items()), len(collector)))
        else:
            print(f'Executing on {nproc} processors')
            merged_results = mmcv.track_parallel_progress(
                merge_func, list(collector.items()), nproc)

        # Return a zipped list of merged results
        return zip(*merged_results)

    def _results2submission(self, id_list, dets_list, out_folder=None):
        """Generate the submission of full images.

        Args:
            id_list (list): Id of images.
            dets_list (list): Detection results of per class.
            out_folder (str, optional): Folder of submission.
        """
        if osp.exists(out_folder):
            raise ValueError(f'The out_folder should be a non-exist path, '
                             f'but {out_folder} is existing')
        os.makedirs(out_folder)

        files = [
            osp.join(out_folder, 'Task1_' + cls + '.txt')
            for cls in self.CLASSES
        ]
        file_objs = [open(f, 'w') for f in files]
        for img_id, dets_per_cls in zip(id_list, dets_list):
            for f, dets in zip(file_objs, dets_per_cls):
                if dets.size == 0:
                    continue
                bboxes = obb2poly_np(dets, self.version)
                for bbox in bboxes:
                    txt_element = [img_id, str(bbox[-1])
                                   ] + [f'{p:.2f}' for p in bbox[:-1]]
                    f.writelines(' '.join(txt_element) + '\n')

        for f in file_objs:
            f.close()

        target_name = osp.split(out_folder)[-1]
        with zipfile.ZipFile(
                osp.join(out_folder, target_name + '.zip'), 'w',
                zipfile.ZIP_DEFLATED) as t:
            for f in files:
                t.write(f, osp.split(f)[-1])

        return files

    def format_results(self, results, submission_dir=None, nproc=4, **kwargs):
        """Format the results to submission text (standard format for DOTA
        evaluation).

        Args:
            results (list): Testing results of the dataset.
            submission_dir (str, optional): The folder that contains submission
                files. If not specified, a temp folder will be created.
                Default: None.
            nproc (int, optional): number of process.

        Returns:
            tuple:

                - result_files (dict): a dict containing the json filepaths
                - tmp_dir (str): the temporal directory created for saving \
                    json files when submission_dir is not specified.
        """
        nproc = min(nproc, os.cpu_count())
        assert isinstance(results, list), 'results must be a list'
        assert len(results) == len(self), (
            f'The length of results is not equal to '
            f'the dataset len: {len(results)} != {len(self)}')
        if submission_dir is None:
            submission_dir = tempfile.TemporaryDirectory()
        else:
            tmp_dir = None

        print('\nMerging patch bboxes into full image!!!')
        start_time = time.time()
        id_list, dets_list = self.merge_det(results, nproc)
        stop_time = time.time()
        print(f'Used time: {(stop_time - start_time):.1f} s')

        result_files = self._results2submission(id_list, dets_list,
                                                submission_dir)

        return result_files, tmp_dir


def _merge_func(info, CLASSES, iou_thr):
    """Merging patch bboxes into full image.

    Args:
        CLASSES (list): Label category.
        iou_thr (float): Threshold of IoU.
    """
    img_id, label_dets = info
    label_dets = np.concatenate(label_dets, axis=0)

    labels, dets = label_dets[:, 0], label_dets[:, 1:]

    big_img_results = []
    for i in range(len(CLASSES)):
        if len(dets[labels == i]) == 0:
            big_img_results.append(dets[labels == i])
        else:
            try:
                cls_dets = torch.from_numpy(dets[labels == i]).cuda()
            except:  # noqa: E722
                cls_dets = torch.from_numpy(dets[labels == i])
            nms_dets, keep_inds = nms_rotated(cls_dets[:, :5], cls_dets[:, -1],
                                              iou_thr)
            big_img_results.append(nms_dets.cpu().numpy())
    return img_id, big_img_results



# Copyright (c) Open-MMLab. All rights reserved.
import os.path as osp
import time
from tempfile import TemporaryDirectory

import torch
from torch.optim import Optimizer

import mmcv
from mmcv.parallel import is_module_wrapper
from mmcv.runner.checkpoint import weights_to_cpu, get_state_dict

try:
    import apex
except:
    print('apex is not installed')


def save_checkpoint(model, filename, optimizer=None, meta=None):
    """Save checkpoint to file.

    The checkpoint will have 4 fields: ``meta``, ``state_dict`` and
    ``optimizer``, ``amp``. By default ``meta`` will contain version
    and time info.

    Args:
        model (Module): Module whose params are to be saved.
        filename (str): Checkpoint filename.
        optimizer (:obj:`Optimizer`, optional): Optimizer to be saved.
        meta (dict, optional): Metadata to be saved in checkpoint.
    """
    if meta is None:
        meta = {}
    elif not isinstance(meta, dict):
        raise TypeError(f'meta must be a dict or None, but got {type(meta)}')
    meta.update(mmcv_version=mmcv.__version__, time=time.asctime())

    if is_module_wrapper(model):
        model = model.module

    if hasattr(model, 'CLASSES') and model.CLASSES is not None:
        # save class name to the meta
        meta.update(CLASSES=model.CLASSES)

    checkpoint = {
        'meta': meta,
        'state_dict': weights_to_cpu(get_state_dict(model))
    }
    # save optimizer state dict in the checkpoint
    if isinstance(optimizer, Optimizer):
        checkpoint['optimizer'] = optimizer.state_dict()
    elif isinstance(optimizer, dict):
        checkpoint['optimizer'] = {}
        for name, optim in optimizer.items():
            checkpoint['optimizer'][name] = optim.state_dict()

    # save amp state dict in the checkpoint
    checkpoint['amp'] = apex.amp.state_dict()

    if filename.startswith('pavi://'):
        try:
            from pavi import modelcloud
            from pavi.exception import NodeNotFoundError
        except ImportError:
            raise ImportError(
                'Please install pavi to load checkpoint from modelcloud.')
        model_path = filename[7:]
        root = modelcloud.Folder()
        model_dir, model_name = osp.split(model_path)
        try:
            model = modelcloud.get(model_dir)
        except NodeNotFoundError:
            model = root.create_training_model(model_dir)
        with TemporaryDirectory() as tmp_dir:
            checkpoint_file = osp.join(tmp_dir, model_name)
            with open(checkpoint_file, 'wb') as f:
                torch.save(checkpoint, f)
                f.flush()
            model.create_file(checkpoint_file, name=model_name)
    else:
        mmcv.mkdir_or_exist(osp.dirname(filename))
        # immediately flush buffer
        with open(filename, 'wb') as f:
            torch.save(checkpoint, f)
            f.flush()



import os 
import os.path as osp 
import glob 
import json 
import cv2
from tqdm import tqdm
from shutil import copyfile
import numpy as np
from visionsuite.utils.helpers import get_filename

from visionsuite.utils.dataset.formats.labelme.utils import get_mask_from_labelme


def labelme2mask(input_dir, output_dir, class2label, input_format, modes, width=None, height=None, vis=False, output_format='bmp'):

    if not osp.exists(output_dir):
        os.makedirs(output_dir)

    for mode in modes:
        _output_dir = osp.join(output_dir, mode)
        if not osp.exists(_output_dir):
            os.mkdir(_output_dir)
            
            
        if vis:
            vis_dir = osp.join(_output_dir, 'vis')
            if not osp.exists(vis_dir):
                os.mkdir(vis_dir)
            
        json_files = glob.glob(osp.join(input_dir, mode, '*.json'))
        
        for json_file in tqdm(json_files, desc=mode):
            img_file = osp.splitext(json_file)[0] + f'.{input_format}'
            filename = get_filename(json_file, False)
            with open(json_file, 'r') as jf:
                anns = json.load(jf)
                
                
            if width is None and height is None:
                width, height = anns['imageWidth'], anns['imageHeight']
            mask = get_mask_from_labelme(json_file, width, height, class2label, 
                                        format='opencv')
            import numpy as np
            _mask = np.unique(mask)
            mask_output_dir = osp.join(_output_dir, 'masks')
            if not osp.exists(mask_output_dir):
                os.mkdir(mask_output_dir)
                        
            image_output_dir = osp.join(_output_dir, 'images')
            if not osp.exists(image_output_dir):
                os.mkdir(image_output_dir)
                
            cv2.imwrite(osp.join(mask_output_dir, filename + f'.{output_format}'), mask)
            copyfile(img_file, osp.join(image_output_dir, filename + f'.{output_format}'))
            
            if vis:
                import numpy as np
                import imgviz 
                
                img = cv2.imread(img_file)
                
                vis_img = np.zeros((height, width*2, 3))
                vis_img[:, :width, :] = img
                color_map = imgviz.label_colormap(50)
                mask = color_map[mask.astype(np.uint8)].astype(np.uint8)
                vis_img[:, width:, :] = mask 
                
                cv2.imwrite(osp.join(vis_dir, filename + '.png'), vis_img)
        
            
if __name__ == '__main__':
    input_dir = '/HDD/datasets/projects/LX/24.12.12/split_labelme_patch_dataset'
    output_dir = '/HDD/datasets/projects/LX/24.12.12/split_mask_patch_dataset'
    # class2label = {'tear': 1}
    # class2label = {'scratch': 1}
    # class2label = {'scratch': 1, 'tear': 2}
    # class2label = {'scratch': 1, 'tear': 2, 'stabbed': 3}
    class2label = {'timber': 1, 'screw': 2}
    modes = ['train', 'val']
    width, height = 1024, 1024
    vis = True
    output_format = 'png'
    input_format = 'png'

    labelme2mask(input_dir, output_dir, class2label, input_format, modes, width, height, vis, output_format)


    # input_dir = '/HDD/datasets/projects/LX/24.11.28_2/datasets_wo_vertical/datasets/tmp_patch_dataset'
    # output_dir = '/HDD/datasets/projects/LX/24.11.28_2/datasets_wo_vertical/datasets/tmp_patch_dataset/vis'
    # # class2label = {'tear': 1}
    # # class2label = {'scratch': 1}
    # # class2label = {'scratch': 1, 'tear': 2}
    # # class2label = {'scratch': 1, 'tear': 2, 'stabbed': 3}
    # class2label = {'timber': 1, 'screw': 2}
    # modes = ['./']
    # width, height = 1024, 1024
    # vis = True
    # output_format = 'png'
    # input_format = 'jpg'

    # labelme2mask(input_dir, output_dir, class2label, input_format, modes, width, height, vis, output_format)





import json
import os.path as osp

def parse_train_log(log_filename):
    
    assert osp.exists(log_filename), ValueError(f'There is no such log-file: {log_filename}')
    
    json_data = []
    with open(log_filename, 'r') as f:
        for line in f:
            json_data.append(json.loads(line))
    
    
    
if __name__ == '__main__':
    log_filename = '/HDD/datasets/projects/rich/24.06.19/benchmark/std/rotated_imted_vb1m_oriented_rcnn_vit_base_1x_dota_ms_rr_le90_stdc_xyawh321v/20240902_055256.log.json'
    parse_train_log(log_filename)


import subprocess

try:
    import wandb as wb
except (ImportError, AssertionError):
    subprocess.run(["pip", "install", "wandb"])
    import wandb as wb

# assert hasattr(wb, '__version__')  # verify package is not directory
from pathlib import Path

import numpy as np
import pandas as pd
from omegaconf import OmegaConf

_processed_plots = {}


def _custom_table(
    x, y, classes, title="Precision Recall Curve", x_title="Recall", y_title="Precision"
):
    """
    Create and log a custom metric visualization to wandb.plot.pr_curve.

    This function crafts a custom metric visualization that mimics the behavior of wandb's default precision-recall curve
    while allowing for enhanced customization. The visual metric is useful for monitoring model performance across different classes.

    Args:
        x (List): Values for the x-axis; expected to have length N.
        y (List): Corresponding values for the y-axis; also expected to have length N.
        classes (List): Labels identifying the class of each point; length N.
        title (str, optional): Title for the plot; defaults to 'Precision Recall Curve'.
        x_title (str, optional): Label for the x-axis; defaults to 'Recall'.
        y_title (str, optional): Label for the y-axis; defaults to 'Precision'.

    Returns:
        (wandb.Object): A wandb object suitable for logging, showcasing the crafted metric visualization.
    """
    df = pd.DataFrame({"class": classes, "y": y, "x": x}).round(3)
    fields = {"x": "x", "y": "y", "class": "class"}
    string_fields = {"title": title, "x-axis-title": x_title, "y-axis-title": y_title}
    return wb.plot_table(
        "wandb/area-under-curve/v0",
        wb.Table(dataframe=df),
        fields=fields,
        string_fields=string_fields,
    )


def _plot_curve(
    x,
    y,
    names=None,
    id="precision-recall",
    title="Precision Recall Curve",
    x_title="Recall",
    y_title="Precision",
    num_x=100,
    only_mean=False,
):
    """
    Log a metric curve visualization.

    This function generates a metric curve based on input data and logs the visualization to wandb.
    The curve can represent aggregated data (mean) or individual class data, depending on the 'only_mean' flag.

    Args:
        x (np.ndarray): Data points for the x-axis with length N.
        y (np.ndarray): Corresponding data points for the y-axis with shape CxN, where C represents the number of classes.
        names (list, optional): Names of the classes corresponding to the y-axis data; length C. Defaults to an empty list.
        id (str, optional): Unique identifier for the logged data in wandb. Defaults to 'precision-recall'.
        title (str, optional): Title for the visualization plot. Defaults to 'Precision Recall Curve'.
        x_title (str, optional): Label for the x-axis. Defaults to 'Recall'.
        y_title (str, optional): Label for the y-axis. Defaults to 'Precision'.
        num_x (int, optional): Number of interpolated data points for visualization. Defaults to 100.
        only_mean (bool, optional): Flag to indicate if only the mean curve should be plotted. Defaults to True.

    Note:
        The function leverages the '_custom_table' function to generate the actual visualization.
    """
    # Create new x
    if names is None:
        names = []
    x_new = np.linspace(x[0], x[-1], num_x).round(5)

    # Create arrays for logging
    x_log = x_new.tolist()
    y_log = np.interp(x_new, x, np.mean(y, axis=0)).round(3).tolist()

    if only_mean:
        table = wb.Table(data=list(zip(x_log, y_log)), columns=[x_title, y_title])
        wb.run.log({title: wb.plot.line(table, x_title, y_title, title=title)})
    else:
        classes = ["mean"] * len(x_log)
        for i, yi in enumerate(y):
            x_log.extend(x_new)  # add new x
            y_log.extend(np.interp(x_new, x, yi))  # interpolate y to new x
            classes.extend([names[i]] * len(x_new))  # add class names
        wb.log(
            {id: _custom_table(x_log, y_log, classes, title, x_title, y_title)},
            commit=False,
        )


def _log_plots(plots, step):
    """Logs plots from the input dictionary if they haven't been logged already at the specified step."""
    for name, params in plots.items():
        timestamp = params["timestamp"]
        if _processed_plots.get(name) != timestamp:
            wb.run.log({name.stem: wb.Image(str(name))}, step=step)
            _processed_plots[name] = timestamp


def on_runner_set_variables(engine):
    try:
        engine.flags.is_wandb_logined = wb.login(
            key="817e3db4693a77186b46ea420d9fcb5bdd74f4d8"
        )
        name = ""
        if hasattr(engine.configs, "db") and hasattr(
            engine.configs.db, "sub_project_name"
        ):
            name += f"{engine.configs.db.sub_project_name}"
        if hasattr(engine.configs, "logging") and hasattr(
            engine.configs.logging, "logs_dir"
        ):
            name += f"_{str(Path(engine.configs.logging.logs_dir).parent.parent).split('/')[-1]}"

        wb.run or wb.init(
            project=engine.configs.db.project_name,
            name=name,
            config=OmegaConf.to_container(engine.configs.__dict__["_configs"]),
        )

        engine.flags.is_wandb_initialized = True

        if engine._callbacks.logger is not None:
            engine._callbacks.logger.info(f"Logined WanDB: ", on_runner_set_variables.__name__)
            engine._callbacks.logger.info(
                f" - project name: {name}", on_runner_set_variables.__name__
            )

    except Exception as error:
        engine.flags.is_wandb_logined, engine.flags.is_wandb_initialized = False, False
        if engine._callbacks.logger is not None:
            engine._callbacks.logger.info(f"CANNOT connect to WanDB b/c {error}")


def on_fit_epoch_end(engine):
    """Logs training metrics and model information at the end of an epoch."""
    if engine.flags.is_wandb_logined and engine.flags.is_wandb_initialized:
        wb.run.log(engine.metrics, step=engine.epoch + 1)
        _log_plots(engine.plots, step=engine.epoch + 1)
        _log_plots(engine.validator.plots, step=engine.epoch + 1)
    # if engine.epoch == 0:
    #     wb.run.log(model_info_for_loggers(engine), step=engine.epoch + 1)


def on_trainer_epoch_end(engine):
    """Log metrics and save images at the end of each training epoch."""
    # wb.run.log(engine.label_loss_items(engine.tloss, prefix='train'), step=engine.epoch + 1)
    # wb.run.log(engine.lr, step=engine.epoch + 1)
    if engine.flags.is_wandb_logined and engine.flags.is_wandb_initialized:
        log = {}
        for key, val in engine.train_epoch_results.log.items():
            log["train_" + key] = val
        wb.log(log)
        if engine._callbacks.logger is not None:
            engine._callbacks.logger.info(
                f"Logged train results of one epoch into WanDB",
                on_trainer_epoch_end.__name__,
            )
        # if engine.epoch == 1:
        #     _log_plots(engine.plots, step=engine.epoch + 1)


def on_train_end(engine):
    """Save the best model as an artifact at end of training."""
    if engine.flags.is_wandb_logined and engine.flags.is_wandb_initialized:
        _log_plots(engine.validator.plots, step=engine.epoch + 1)
        _log_plots(engine.plots, step=engine.epoch + 1)
        art = wb.Artifact(type="model", name=f"run_{wb.run.id}_model")
        if engine.best.exists():
            art.add_file(engine.best)
            wb.run.log_artifact(art, aliases=["best"])
        for curve_name, curve_values in zip(
            engine.validator.metrics.curves, engine.validator.metrics.curves_results
        ):
            x, y, x_title, y_title = curve_values
            _plot_curve(
                x,
                y,
                names=list(engine.validator.metrics.names.values()),
                id=f"curves/{curve_name}",
                title=curve_name,
                x_title=x_title,
                y_title=y_title,
            )
        wb.run.finish()  # required or run continues on dashboard
        if engine._callbacks.logger is not None:
            engine._callbacks.logger.info(
                f"Logged final train results of one epoch into WanDB",
                on_train_end.__name__,
            )


def on_val_end(engine):
    if engine.flags.is_wandb_logined and engine.flags.is_wandb_initialized:
        log = {}
        for key, val in engine.val_results.log.items():
            log["val_" + key] = val
        wb.log(log)

        if engine._callbacks.logger is not None:
            engine._callbacks.logger.info(
                f"Logged val. results of one epoch into WanDB", on_val_end.__name__
            )


callbacks = (
    {
        "on_runner_set_variables": on_runner_set_variables,
        "on_trainer_epoch_end": on_trainer_epoch_end,
        "on_val_end": on_val_end,
        "on_fit_epoch_end": on_fit_epoch_end,
        "on_train_end": on_train_end,
    }
    if wb
    else {}
)



import yaml
import os.path as osp
from visionsuite.engines.utils.helpers import create_output_dir, mkdir
from visionsuite.engines.utils.loggers.monitor import Monitor
from visionsuite.engines.utils.bases.base_oop_module import BaseOOPModule


class Archive(BaseOOPModule):
    _train_dirs = ['val', 'cfgs', 'logs', 'weights']
    _test_dirs = ['cfgs', 'logs', 'vis']

    def __init__(self, mode, output_dir=None):
        super().__init__(__class__.__name__)
        self.args = None
        self._mode = mode 
        self._output_dir = output_dir
        
    @property 
    def mode(self):
        return self._mode 
            
    @property
    def output_dir(self):
        return self._output_dir
    
    def build(self, *args, **kwargs):
        super().build(*args, **kwargs)

        self._output_dir = create_output_dir(self.args['output_dir'], mode=self._mode, make_dirs=True)
        assert osp.exists(self._output_dir), RuntimeError(f"Output dir ({self._output_dir}) has not been created")
        
        self.create_directories()
        self.set_logger(log_stream_level=self.args['logger']['log_stream_level'],
                        log_file_level=self.args['logger']['log_file_level'],
                        log_dir=self.logs_dir)        
        self._load()
        
    @BaseOOPModule.track_status
    def _load(self):
        self._load_monitor()
    
    @BaseOOPModule.track_status
    def _load_monitor(self):
        if self.args['monitor']['use']:
            self.monitor = Monitor()
            self.monitor.set(output_dir=self.logs_dir, fn='monitor')
            self.log_info(f"Monitor is LOADED and SET", self._load_monitor.__name__, __class__.__name__)

        else:
            self.monitor = None
            self.log_info(f"Monitor is NOT LOADED", self._load_monitor.__name__, __class__.__name__)
    
    @BaseOOPModule.track_status
    def save_args(self, args):
        with open(osp.join(self.cfgs_dir, 'args.yaml'), 'w') as yf:
            yaml.dump(args, yf, default_flow_style=False)
            
        self.log_info(f"Saved args at {osp.join(self.cfgs_dir, 'args.yaml')}", self.save_args.__name__, __class__.__name__)
    
    @BaseOOPModule.track_status
    def create_directories(self):
        for directory in getattr(self, f'_{self._mode}_dirs'):
            mkdir(osp.join(self._output_dir, directory))
            setattr(self, directory + '_dir', osp.join(self._output_dir, directory))    
    
            self.log_info(f"Created directory: {osp.join(self._output_dir, directory)}", self.create_directories.__name__, __class__.__name__)

        
        
            
    


from visionsuite.engines.utils.bases import BaseTrainLoop


class Loop(BaseTrainLoop):
    
    def build(self, _model, _dataset, _archive=None, *args, **kwargs):
        super().build(_model=_model, _dataset=_dataset, _archive=_archive, *args, **kwargs)
        self.run_callbacks('on_loop_build_start')

        if 'dataloader' in self.args and self.args['dataloader'] is not None:
            from visionsuite.engines.classification.src.dataloaders.build import build_dataloader
            self.train_dataloader = build_dataloader(dataset=self.dataset, mode='train', 
                                                    **self.args['dataloader'], 
                                                    augment=self.args['augment'] if 'augment' in self.args else None)
            self.log_info(f"BUILT train_dataloader: {self.train_dataloader}", self.build.__name__, __class__.__name__)
            
            self.val_dataloader = build_dataloader(dataset=self.dataset, mode='val', 
                                                    **self.args['dataloader'], 
                                                    augment=self.args['augment'] if 'augment' in self.args else None)
            self.log_info(f"BUILT val_dataloader: {self.val_dataloader}", self.build.__name__, __class__.__name__)    
        else:
            self.log_warning(f"NO dataloader", self.build.__name__, __class__.__name__)    

        if 'loss' in self.args and self.args['loss'] is not None:
            from visionsuite.engines.classification.src.losses.build import build_loss
            self.loss = build_loss(**self.args['loss'])
            self.log_info(f"BUILT loss: {self.loss}", self.build.__name__, __class__.__name__)  
        else:
            self.log_warning(f"NO loss", self.build.__name__, __class__.__name__)  
        
        if 'optimizer' in self.args and self.args['optimizer'] is not None:
            from visionsuite.engines.classification.src.optimizers.build import build_optimizer
            self.optimizer = build_optimizer(model=self.model.model, **self.args['optimizer'])
            self.log_info(f"BUILT optimizer: {self.optimizer}", self.build.__name__, __class__.__name__)   
        else:
            self.log_warning(f"NO optimizer", self.build.__name__, __class__.__name__)   
        
        if 'scheduler' in self.args and self.args['scheduler'] is not None:
            from visionsuite.engines.classification.src.schedulers.build import build_scheduler
            self.lr_scheduler = build_scheduler(optimizer=self.optimizer, epochs=self.args['train']['epochs'], **self.args['scheduler'])
            self.log_info(f"BUILT lr_scheduler: {self.lr_scheduler}", self.build.__name__, __class__.__name__)     
        else:
            self.log_warning(f"NO lr_scheduler", self.build.__name__, __class__.__name__)     

        
        if 'train' in self.args and 'trainer' in self.args['train'] and self.args['train']['trainer'] is not None:
            from visionsuite.engines.classification.src.trainers.build import build_trainer
            trainer = build_trainer(**self.args['train']['trainer'])()
            trainer.build(model=self.model, loss=self.loss, optimizer=self.optimizer, 
                                lr_scheduler=self.lr_scheduler, dataloader=self.train_dataloader, 
                                scaler=self.scaler, archive=self.archive,
                                **self.args['train'],
                                _logger=self.args['trainer'].get('logger', None))
            self._loops.append(trainer)
            self.log_info(f"BUILT trainer: {self.trainer}", self.build.__name__, __class__.__name__)
        else:
            self.log_warning(f"NO trainer", self.build.__name__, __class__.__name__)
        
        if 'val' in self.args and 'validator' in self.args['val'] and self.args['val']['validator'] is not None:
            from visionsuite.engines.classification.src.validators.build import build_validator
            validator = build_validator(**self.args['val']['validator'])()
            validator.build(model=self.model, loss=self.loss, dataloader=self.val_dataloader,
                                label2index=self.dataset.label2index, 
                                device=self.args['train']['device'], topk=self.args['train']['topk'],
                                archive=self.archive, **self.args['val'],
                                _logger=self.args['validator'].get('logger', None))
            self._loops.append(validator)
            self.log_info(f"BUILT validator: {self.validator}", self.build.__name__, __class__.__name__)
        else:
            self.log_warning(f"NO validator", self.build.__name__, __class__.__name__)
            
        self._set_resume()
    



########################################################
#                                                      #
#       author: omitted for anonymous submission       #
#                                                      #
#     credits and copyright coming upon publication    #
#                                                      #
########################################################


import argparse


class ArgumentParser(argparse.ArgumentParser):
    def set_common_args(self):
        # paths
        self.add_argument("--results_dir", default="/HDD/etc/contmav/outputs")
        self.add_argument(
            "--last_ckpt",
            default="/HDD/etc/contmav/outputs/cityscapes/checkpoint/model1/ckpt_latest.pth",
            type=str,
            metavar="PATH",
            help="path to latest checkpoint",
        )
        self.add_argument(
            "--load_weights",
            # default="/HDD/etc/contmav/outputs/cityscapes/checkpoint/16_10_2024-19_46_36-476538/best_miou.pth",
            default="",
            type=str,
            metavar="PATH",
            help="path to model weights",
        )
        self.add_argument(
            "--pretrained_dir",
            default="./trained_models/imagenet",
            help="path to the pretrained resnets with different" "encoder block",
        )
        self.add_argument(
            "--id",
            default="checkpoint",
            help="path to the pretrained resnets with different" "encoder block",
        )

        # pretraining
        self.add_argument(
            "--no_imagenet_pretraining",
            dest="pretrained_on_imagenet",
            default=False,
            action="store_false",
            help="Encoder will be initialized randomly. "
            "(If not set encoder will be initialized with "
            "weights pretrained on ImageNet)",
        )
        self.add_argument(
            "--finetune",
            default=None,
            type=str,
            help="path to the weights you want to finetune on.",
        )
        self.add_argument(
            "--freeze",
            default=0,
            type=int,
            help="number of epochs the whole model (except for "
            "the output layer(s) are freezed. Might be "
            "useful when using a pretrained "
            "model on scenenet.",
        )

        # input dimensions
        self.add_argument(
            "--batch_size", type=int, default=4, help="batch size for training"
        )
        self.add_argument(
            "--batch_size_valid",
            type=int,
            default=None,
            help="batch size for validation. Can be typically "
            "2-3 times as large as the batch size for "
            "training. If None it will be the same as "
            "--batch_size.",
        )
        self.add_argument(
            "--height",
            type=int,
            default=512,
            help="height of the training images. "
            "Images will be resized to this height.",
        )
        self.add_argument(
            "--width",
            type=int,
            default=1024,
            help="width of the training images. "
            "Images will be resized to this width.",
        )

        # epochs
        self.add_argument(
            "--epochs",
            default=500,
            type=int,
            metavar="N",
            help="number of total epochs to run",
        )

        # training hyper parameters
        self.add_argument(
            "--lr",
            "--learning-rate",
            default=0.0001,
            type=float,
            help="maximum learning rate. When using one_cycle "
            "as --lr_scheduler lr will first increase to "
            "the value provided and then slowly decrease.",
        )
        self.add_argument(
            "--weight_decay", "--wd", default=1e-4, type=float, help="weight decay"
        )
        self.add_argument(
            "--momentum", default=0.9, type=float, metavar="M", help="momentum"
        )
        self.add_argument(
            "--optimizer", type=str, default="Adam", choices=["SGD", "Adam"]
        )
        self.add_argument(
            "--class_weighting",
            type=str,
            default="None",
            choices=["median_frequency", "logarithmic", "None"],
            help="which weighting mode to use for weighting the "
            "classes of the unbalanced dataset"
            "for the loss function during training.",
        )
        self.add_argument(
            "--c_for_logarithmic_weighting",
            type=str,
            default=1.02,
            help="the value for restricting the class weights. "
            "The value is only used when --class_weighting "
            "is set to logarithmic",
        )
        self.add_argument(
            "--he_init",
            dest="he_init",
            default=True,
            action="store_true",
            help="Set this if you want to initialize "
            "convolution layers with He initialization.",
        )
        self.add_argument(
            "--valid_full_res",
            default=True,
            action="store_true",
            help="Whether to validate on the full resolution " "(for cityscapes).",
        )

        # model
        self.add_argument(
            "--activation",
            type=str,
            default="relu",
            choices=["relu", "swish", "hswish"],
            help="Which activation function to use in the model",
        )
        self.add_argument(
            "--encoder",
            type=str,
            default="resnet34",
            choices=["resnet18", "resnet34", "resnet50"],
            help="Wich encoder to use for rgb features.",
        )
        self.add_argument(
            "--encoder_block",
            type=str,
            default="NonBottleneck1D",
            choices=["BasicBlock", "NonBottleneck1D"],
            help="The block that is used in the ResNet encoder."
            "The NonBottleneck1D achieves better "
            "results than the BasicBlock.",
        )
        self.add_argument(
            "--nr_decoder_blocks",
            type=int,
            default=[3],
            nargs="+",
            help="How many decoder blocks are used in each "
            "decoder module. This variable is only used "
            'when decoder_block != "None"',
        )
        self.add_argument(
            "--modality",
            type=str,
            default="rgb",
        )
        self.add_argument(
            "--encoder_decoder_fusion",
            type=str,
            default="add",
            choices=["add", "None"],
            help="How to fuse encoder feature maps into the "
            "decoder. If None no encoder feature maps are "
            "fused into the decoder.",
        )
        self.add_argument(
            "--context_module",
            type=str,
            default="appm-1-2-4-8",
            choices=["ppm", "None", "ppm-1-2-4-8", "appm", "appm-1-2-4-8"],
            help="Which context module to use.",
        )
        self.add_argument(
            "--channels_decoder",
            type=int,
            default=128,
            help="How many feature maps to use in the decoder. "
            "This is only used when you set:"
            "--decoder_channels_mode constant",
        )
        self.add_argument(
            "--decoder_channels_mode",
            default="decreasing",
            choices=["constant", "decreasing"],
            help="constant: the number of channels in the "
            "decoder stays the same."
            "decreasing: the channel number is decreasing "
            "as the resolution is increasing. Note that "
            "than the argument --channels_decoder is "
            "ignored.",
        )
        self.add_argument(
            "--upsampling",
            default="learned-3x3-zeropad",
            choices=["nearest", "bilinear", "learned-3x3", "learned-3x3-zeropad"],
            help="How to usample in the decoder. "
            "Bilinear upsampling can cause problems"
            "with conversion to TensorRT. learned-3x3 "
            "mimics a bilinear interpolation with nearest "
            "neighbor interpolation and a 3x3 conv "
            "afterwards",
        )

        # dataset
        self.add_argument(
            "--dataset",
            default="cityscapes",
            choices=[
                "cityscapes",
            ],
        )
        self.add_argument(
            "--dataset_dir",
            default="/HDD/datasets/public/cityscapes",
            help="Path to dataset root.",
        )
        self.add_argument(
            "--aug_scale_min",
            default=0.5,
            type=float,
            help="the minimum scale for random rescaling the " "training data.",
        )
        self.add_argument(
            "--aug_scale_max",
            default=2.0,
            type=float,
            help="the maximum scale for random rescaling the " "training data.",
        )
        self.add_argument(
            "--overfit",
            default=False,
            type=bool,
            help="tests the overfit on one image",
        )
        self.add_argument(
            "--num_classes",
            default=19,
            type=int,
            help="tests the overfit on one image",
        )

        # losses
        self.add_argument(
            "--obj",
            default=False,
            type=bool,
            help="use the objectosphere loss",
        )
        self.add_argument(
            "--mav",
            default=False,
            type=bool,
            help="use the MAV loss",
        )
        self.add_argument(
            "--closs",
            default=False,
            type=bool,
            help="use contrastive loss",
        )
        self.add_argument(
            "--loss_weights",
            default=[1, 1, 1, 1],
            type=bool,
            help="use the MAV loss",
        )

        # others
        self.add_argument(
            "-j",
            "--workers",
            default=10,
            type=int,
            metavar="N",
            help="number of data loading workers",
        )
        self.add_argument(
            "--debug",
            default=False,
            action="store_true",
            help="Only one batch in training and validation.",
        )



from pathlib import Path 
FILE = Path(__file__).resolve()
ROOT = FILE.parent

from abc import abstractmethod
import os.path as osp
import warnings 

from visionsuite.engines.utils.helpers import yaml2dict, update_dict
from visionsuite.engines.utils.torch_utils.utils import parse_device_ids, set_torch_deterministic, get_device
from visionsuite.engines.utils.torch_utils.dist import init_distributed_mode
from visionsuite.engines.utils.archives import Archive
from visionsuite.engines.utils.loggers import Logger

            
class BaseRunner(Logger):
    def __init__(self, task, name=None):
        super().__init__(name=name)
        self._task = task
        self._mode = None
        self.pipelines = {}
        
    @property 
    def task(self):
        return self._task 
    
    @property 
    def mode(self):
        return self._mode 
    
    @abstractmethod
    def set_configs(self, mode, cfgs_file=None, default_cfgs_file=None, *args, **kwargs):
        self._mode = mode
        if cfgs_file and osp.exists(cfgs_file):
            cfgs = yaml2dict(cfgs_file)
        else:
            cfgs = {}
            warnings.warn(f'There is no such cfgs file: {cfgs_file}')

        def _parse_args(args):
            # device_ids
            args[mode]['device_ids'] = parse_device_ids(args[mode]['device_ids'])
                        
        if default_cfgs_file is None:
            default_cfgs_file=ROOT.parents[1] / self._task / f'cfgs/default.yaml'
        default_cfgs = yaml2dict(default_cfgs_file)

        update_dict(default_cfgs, cfgs)  
        update_dict(default_cfgs, kwargs)          
        
        self.args = default_cfgs
        _parse_args(self.args)


    @abstractmethod
    def set_variables(self):
        # logger
        for key in ['archive']:
            if key not in self.args:
                self.args[key] = {'logger': {}}
            
            if 'logger' not in self.args[key]:
                self.args[key]['logger'] = self.args['logger']
            else:
                for key2, val2 in self.args['logger'].items():
                    if key2 not in self.args[key]['logger']:
                        self.args[key]['logger'][key2] = val2
        
        self._archive = Archive(self.mode)
        self._archive.build(**self.args['archive'])
        self._archive.save_args(self.args)
        
        # logger
        for key in ['runner', 'model', 'loop', 'trainer', 'validator', 'tester', 'dataset']:
            if key not in self.args:
                self.args[key] = {'logger': {}}
            
            if 'logger' not in self.args[key]:
                self.args[key]['logger'] = self.args['logger']
                self.args[key]['logger']['logs_dir'] = self._archive.logs_dir
            else:
                for key2, val2 in self.args['logger'].items():
                    if key2 not in self.args[key]['logger']:
                        self.args[key]['logger'][key2] = val2
                self.args[key]['logger']['logs_dir'] = self._archive.logs_dir
                
        self.set_logger(log_stream_level=self.args['runner']['logger']['log_stream_level'],
                        log_file_level=self.args['runner']['logger']['log_file_level'],
                        log_dir=self.args['runner']['logger']['logs_dir']) 
        
        
        self.log_info(f"Args: {self.args}", self.set_variables.__name__, __class__.__name__)
        
        init_distributed_mode(self.args, self._mode)
        self.log_info(f"Initialize distribution mode: {self.args['distributed']['use']}", self.set_variables.__name__, __class__.__name__)
        
        set_torch_deterministic(self.args[self._mode]['use_deterministic_algorithms'])
        self.log_info(f"Set torch deterministic: {self.args[self._mode]['use_deterministic_algorithms']}", self.set_variables.__name__, __class__.__name__)

        self.args[self._mode]['device'] = get_device(self.args[self._mode]['device'])
        self.log_info(f"Set devices: {self.args[self._mode]['device']}", self.set_variables.__name__, __class__.__name__)
        
        
    @abstractmethod
    def run(self):
        pass 
    



from .base_engine import BaseEngine 
from .base_train_runner import BaseTrainRunner
from .base_test_runner import BaseTestRunner
from .base_oop_module import BaseOOPModule
from .base_results import BaseResults
from .base_train_loop import BaseTrainLoop
from .base_runner import BaseRunner
from .base_test_loop import BaseTestLoop

__all__ = ['BaseEngine', 'BaseTrainRunner', 'BaseTestRunner', 'BaseOOPModule', 'BaseResults', 'BaseTrainLoop', 'BaseRunner', 'BaseTestLoop']


from visionsuite.engines.segmentation.utils.registry import OPTIMIZERS
    

def build_optimizer(model, **config):
    
    from visionsuite.engines.utils.helpers import get_params_from_obj
    optim_obj = OPTIMIZERS.get(config['type'])
    optim_params = get_params_from_obj(optim_obj)
    for key in optim_params.keys():
        if key in config:
            optim_params[key] = config[key]
            
        if key == 'params':
            model.apply_aux_loss_to_params_to_optimize(config['lr'])
            optim_params[key] = model.params_to_optimize
        
    return optim_obj(**optim_params)
    



import os.path as osp

from visionsuite.engines.utils.loggers import Logger
from omegaconf.listconfig import ListConfig


class BaseSlicer:

    def __init__(self, child=None):
        # input
        self._child = child
        self.logger = None

        self.__mode = None
        self.__input_dir = None
        self.__exts = []
        self.__classes = []
        self._roi_info = None
        self._patch_info = None
        self._roi_from_json = False

        # output
        self.__num_data = 0
        self.__img_files = []
        self.__imgs_info = []  # final output
        """
        - imgs_info: [
                {
                    'img_file' : image file name, # string
                    'patches': [
                                [x1, y1, x2, y2],
                                [x1, y1, x2, y2],
                                ...
                            ], # list of lists
                    'counts': [0, 0, ...], # list
                    'backgrounds': [[x1, y1, x2, y2], [x1, y1, x2, y2], ...], # list of lists
                    'labels': [
                                [{'label': 'abc', 'points': [[x1, y1], [x2, y2], ...], 'shape_type': '', 'roi': [x1, y1, x2, y2]},
                                 {'label': '', 'points': [[x1, y1], [x2, y2], ...], 'shape_type': '', 'roi': [x1, y1, x2, y2]}, ...],
                                [{'label': '', 'points': [[x1, y1], [x2, y2], ...], 'shape_type': '', 'roi': [x1, y1, x2, y2]},
                                 {'label': '', 'points': [[x1, y1], [x2, y2], ...], 'shape_type': '', 'roi': [x1, y1, x2, y2]}, ...],
                                ...
                            ] # list of lists
                },
                {
                    'img_file' : image file name,
                    'patches': ...,
                    ...
                },
                ...
        ]
            - each patch coordinate is synched to each labels list and count.
        - backgrounds exist only when using patch-based training
    """

    @property
    def mode(self):
        return self.__mode

    @mode.setter
    def mode(self, value):
        assert isinstance(value, str), ValueError(
            f"Mode variable must be string, not {type(value)}"
        )
        self.__mode = value

    @property
    def input_dir(self):
        return self.__input_dir

    @input_dir.setter
    def input_dir(self, value):
        assert isinstance(value, str), ValueError(
            f"Input directory variable must be string, not {type(value)}"
        )
        assert osp.exists(value), ValueError(f"There is no such directory: {value}")

        self.__input_dir = value

    @property
    def exts(self):
        return self.__exts

    @exts.setter
    def exts(self, value):
        assert isinstance(value, str) or (
            isinstance(value, (list, ListConfig))
            and len(value) != 0
            and isinstance(value[0], str)
        ), ValueError(
            f"Image extension variable must be string or list of string, not {type(value)}"
        )

        self.__exts = value

    @property
    def classes(self):
        return self.__classes

    @classes.setter
    def classes(self, value):
        assert (
            isinstance(value, (list, ListConfig))
            and len(value) != 0
            and isinstance(value[0], str)
        ), ValueError(f"classes variable must be list of string, not {type(value)}")
        self.__classes = value

    @property
    def imgs_info(self):
        return self.__imgs_info

    @imgs_info.setter
    def imgs_info(self, value):
        # assert value > 0, ValueError(f"Number of data must be higher than 0, not {value}")
        self.__imgs_info.append(value)

    @property
    def roi_info(self):
        return self._roi_info

    @roi_info.setter
    def roi_info(self, value):
        # assert value > 0, ValueError(f"Number of data must be higher than 0, not {value}")
        self._roi_info = value

    def set_log(self, log_dir=None, log_stream_level="DEBUG", log_file_level="DEBUG"):

        if self._child is not None:
            log_name = self._child
        else:
            log_name = None
        self._logger = Logger(name=log_name)
        self._logger.set(
            log_dir=log_dir,
            log_stream_level=log_stream_level,
            log_file_level=log_file_level,
        )



from visionsuite.engines.classification.utils.registry import SCHEDULERS


def build_scheduler(optimizer, epochs, **config):
    from visionsuite.engines.utils.helpers import get_params_from_obj
    if config['type'] == "CosineAnnealingLR":
        warmup_total_iters = 0 
        if config['warmup_scheduler'] and 'total_iters' in config['warmup_scheduler'] and config['warmup_scheduler']['total_iters'] > 0:
            warmup_total_iters = config['warmup_scheduler']['total_iters']
        main_lr_scheduler = SCHEDULERS.get(config['type'], case_sensitive=config['case_sensitive'])(
            optimizer, T_max = epochs - warmup_total_iters, eta_min=config['lr_min']
        )
    else:
        scheduler_obj = SCHEDULERS.get(config['type'], case_sensitive=config['case_sensitive'])
        if not scheduler_obj:
            raise RuntimeError(
                f"Invalid lr scheduler '{config['type']}'")
        scheduler_params = get_params_from_obj(scheduler_obj)
        for key in scheduler_params.keys():
            if key in config:
                scheduler_params[key] = config[key]
                
            if key == 'optimizer':
                scheduler_params[key] = optimizer
                
        main_lr_scheduler = scheduler_obj(**scheduler_params)
    
    if config['warmup_scheduler'] and config['warmup_scheduler']['total_iters'] > 0:
        scheduler_obj = SCHEDULERS.get(config['warmup_scheduler']['type'], case_sensitive=config['warmup_scheduler']['case_sensitive'])
        if not scheduler_obj:
            raise RuntimeError(
                f"Invalid lr scheduler '{config['warmup_scheduler']['type']}'. Only ConstantLR and LinearLR are available")
        scheduler_params = get_params_from_obj(scheduler_obj)
        for key in scheduler_params.keys():
            if key in config['warmup_scheduler']:
                scheduler_params[key] = config['warmup_scheduler'][key]
                
            if key == 'optimizer':
                scheduler_params[key] = optimizer
                
        warmup_lr_scheduler = scheduler_obj(**scheduler_params)
        lr_scheduler = SCHEDULERS.get(config['warmup_scheduler']['integrate_scheduler']['type'],
                                      case_sensitive=config['warmup_scheduler']['integrate_scheduler']['case_sensitive'])(
            optimizer, schedulers=[warmup_lr_scheduler, main_lr_scheduler], milestones=[config['warmup_scheduler']['total_iters']]
        )
    else:
        lr_scheduler = main_lr_scheduler
        
    return lr_scheduler

if __name__ == '__main__':
    import torch 
    import matplotlib.pyplot as plt

    # Example configuration for the scheduler
    config = {
        'type': 'StepLR',
        'case_sensitive': False,
        'eta_min': 0.001,
        'gamma': 0.1,
        'step_size': 30,
        'warmup_scheduler': {
            'type': 'LinearLR',
            'total_iters': 10,
            'start_factor': 0.1,
            'case_sensitive': False,
            'integrate_scheduler': {
                'type': 'SequentialLR',
                'case_sensitive': False
            }
        }
    }

    # Initialize optimizer and scheduler
    optimizer = torch.optim.SGD([torch.tensor(1.0)], lr=0.1)
    epochs = 100
    scheduler = build_scheduler(optimizer, epochs, config)

    # Track learning rate for each epoch
    lrs = []
    for epoch in range(epochs):
        optimizer.step()
        lrs.append(optimizer.param_groups[0]['lr'])
        scheduler.step()

    # Plot the learning rate schedule
    plt.figure(figsize=(10, 5))
    plt.plot(range(epochs), lrs, label="Learning Rate")
    plt.xlabel("Epoch")
    plt.ylabel("Learning Rate")
    plt.title("Learning Rate Schedule over Epochs")
    plt.legend()
    plt.grid(True)
    plt.show()


import os.path as osp
import glob 
import json

def preds2metrics(preds_json, class2idx):
    with open(preds_json, 'r') as jf:
        anns = json.load(jf)
    
    detections = []
    for filename, ann in anns.items():
        for _class, val in ann['idx2xyxys'].items():
            if 'bbox' in val:
                for box, conf in zip(val['bbox'], val['confidence']):
                    if ann['idx2class'][_class] not in class2idx:
                        class2idx[ann['idx2class'][_class]] = len(class2idx)
                    detections.append([filename, int(class2idx[ann['idx2class'][_class]]), float(conf), (box[0][0], box[0][1], box[1][0], box[1][1])])
            elif 'polygon' in val:
                if 'confidence' in val:
                    for box, conf in zip(val['polygon'], val['confidence']):
                        if ann['idx2class'][_class] not in class2idx:
                            class2idx[ann['idx2class'][_class]] = len(class2idx)
                        detections.append([filename, int(class2idx[ann['idx2class'][_class]]), float(conf), tuple([_point for __point in box for _point in __point])])
                else:
                    for polygons in zip(val['polygon']):
                        if ann['idx2class'][_class] not in class2idx:
                            class2idx[ann['idx2class'][_class]] = len(class2idx)
                            
                        # dets = []
                        # for polygon in polygons:
                        #     for points in polygon:
                        #         dets.append(points[0])
                        #         dets.append(points[1])
                                
                        detections.append([filename, int(class2idx[ann['idx2class'][_class]]), None, tuple(coord for polygon in polygons for points in polygon for coord in points)])


                    
    return detections, class2idx
        
    
if __name__ == '__main__':
    preds_json = '/DeepLearning/_projects/sungjin_body/tests/yolov8_patch_v2/winter/w_json/학습/preds.json'
        
    detections = preds2metrics(preds_json)
    print(detections)


import os.path as osp 
import glob 
import cv2 
import numpy as np

input_dir = '/HDD/datasets/projects/LX/24.11.28_2/datasets_wo_vertical/datasets/split_mask_patch_dataset/val/masks'

img_files = glob.glob(osp.join(input_dir, '*.jpg'))

for img_file in img_files:
    img = cv2.imread(img_file, 0)
    
    _img = np.unique(img)
    if 3. in list(_img):
        print(_img)
        print(img_file)
    




#!/usr/bin/env python

import warnings

import torch
import torch.nn as nn

from mmcv.cnn import build_conv_layer, build_norm_layer
from mmcv.runner import BaseModule, Sequential
from mmcv.utils.parrots_wrapper import _BatchNorm

from mmseg.models.backbones.resnet import BasicBlock, Bottleneck
from mmseg.models.backbones.hrnet import HRModule
from mmseg.ops import Upsample, resize

from ..builder import BACKBONES


class ModHRModule(HRModule):
    """High-Resolution Module for HRNet.

    In this module, every branch has 4 BasicBlocks/Bottlenecks. Fusion/Exchange
    is in this module.
    """

    def __init__(
        self,
        down_strides=(2, 2, 2),
        **kwargs,
    ):
        self.down_strides = down_strides
        super().__init__(**kwargs)

    def _make_fuse_layers(self):
        """Build fuse layer."""
        if self.num_branches == 1:
            return None

        num_branches = self.num_branches
        in_channels = self.in_channels
        fuse_layers = []
        num_out_branches = num_branches if self.multiscale_output else 1
        for i in range(num_out_branches):
            fuse_layer = []
            for j in range(num_branches):
                if j > i:
                    # we need to upsample j to i
                    # need to calculate the upsampling scale

                    scale = 1
                    for k in range(i, j):
                        scale *= self.down_strides[k]

                    fuse_layer.append(
                        nn.Sequential(
                            build_conv_layer(
                                self.conv_cfg,
                                in_channels[j],
                                in_channels[i],
                                kernel_size=1,
                                stride=1,
                                padding=0,
                                bias=False,
                            ),
                            build_norm_layer(self.norm_cfg, in_channels[i])[1],
                            # we set align_corners=False for HRNet
                            Upsample(
                                scale_factor=scale,
                                mode="bilinear",
                                align_corners=False,
                            ),
                        )
                    )
                elif j == i:
                    # same resolution
                    fuse_layer.append(None)
                else:
                    # we need to downsample j to i
                    # employ multiple conv layers with strides (2 to downsample x2)
                    conv_downsamples = []
                    for k in range(i - j):
                        if k == i - j - 1:
                            # final
                            conv_downsamples.append(
                                nn.Sequential(
                                    build_conv_layer(
                                        self.conv_cfg,
                                        in_channels[j],
                                        in_channels[i],
                                        kernel_size=3,
                                        stride=self.down_strides[j + k],
                                        padding=1,
                                        bias=False,
                                    ),
                                    build_norm_layer(self.norm_cfg, in_channels[i])[1],
                                )
                            )
                        else:
                            conv_downsamples.append(
                                nn.Sequential(
                                    build_conv_layer(
                                        self.conv_cfg,
                                        in_channels[j],
                                        in_channels[j],
                                        kernel_size=3,
                                        stride=self.down_strides[j + k],
                                        padding=1,
                                        bias=False,
                                    ),
                                    build_norm_layer(self.norm_cfg, in_channels[j])[1],
                                    nn.ReLU(inplace=False),
                                )
                            )
                    fuse_layer.append(nn.Sequential(*conv_downsamples))
            fuse_layers.append(nn.ModuleList(fuse_layer))

        return nn.ModuleList(fuse_layers)


@BACKBONES.register_module()
class ModHRNet(BaseModule):
    """Modified HRNet backbone."""

    blocks_dict = {"BASIC": BasicBlock, "BOTTLENECK": Bottleneck}

    def __init__(
        self,
        extra,
        in_channels=3,
        stem_channels=64,
        strides=(2, 2, 2, 2, 2),
        return_stem=False,
        return_sides=False,
        align_corners=False,
        interpolation="bilinear",
        resize_concat_sides=True,
        conv_cfg=None,
        norm_cfg=dict(type="BN", requires_grad=True),
        norm_eval=False,
        with_cp=False,
        frozen_stages=-1,
        zero_init_residual=False,
        multiscale_output=True,
        pretrained=None,
        init_cfg=None,
    ):
        super(ModHRNet, self).__init__(init_cfg)

        self.pretrained = pretrained
        self.zero_init_residual = zero_init_residual
        assert not (
            init_cfg and pretrained
        ), "init_cfg and pretrained cannot be setting at the same time"
        if isinstance(pretrained, str):
            warnings.warn(
                "DeprecationWarning: pretrained is deprecated, "
                'please use "init_cfg" instead'
            )
            self.init_cfg = dict(type="Pretrained", checkpoint=pretrained)
        elif pretrained is None:
            if init_cfg is None:
                self.init_cfg = [
                    dict(type="Kaiming", layer="Conv2d"),
                    dict(type="Constant", val=1, layer=["_BatchNorm", "GroupNorm"]),
                ]
        else:
            raise TypeError("pretrained must be a str or None")

        # Assert configurations of 4 stages are in extra
        assert (
            "stage1" in extra
            and "stage2" in extra
            and "stage3" in extra
            and "stage4" in extra
        )
        # Assert whether the length of `num_blocks` and `num_channels` are
        # equal to `num_branches`
        for i in range(4):
            cfg = extra[f"stage{i + 1}"]
            assert (
                len(cfg["num_blocks"]) == cfg["num_branches"]
                and len(cfg["num_channels"]) == cfg["num_branches"]
            )

        self.extra = extra
        self.conv_cfg = conv_cfg
        self.norm_cfg = norm_cfg
        self.norm_eval = norm_eval
        self.with_cp = with_cp
        self.frozen_stages = frozen_stages

        self.return_stem = return_stem
        self.return_sides = return_sides
        self.interp = interpolation
        self.align_corners = align_corners
        if isinstance(resize_concat_sides, bool):
            self.resize_concat_sides = [resize_concat_sides] * 4
        else:
            self.resize_concat_sides = resize_concat_sides
        assert len(self.resize_concat_sides) == 4

        # stem net
        self.norm1_name, norm1 = build_norm_layer(
            self.norm_cfg, stem_channels, postfix=1
        )
        self.norm2_name, norm2 = build_norm_layer(
            self.norm_cfg, stem_channels, postfix=2
        )
        self.conv1 = build_conv_layer(
            self.conv_cfg,
            in_channels,
            stem_channels,
            kernel_size=3,
            stride=strides.pop(0),
            padding=1,
            bias=False,
        )
        self.add_module(self.norm1_name, norm1)
        self.conv2 = build_conv_layer(
            self.conv_cfg,
            stem_channels,
            stem_channels,
            kernel_size=3,
            stride=strides.pop(0),  # HACK: pop the first one
            padding=1,
            bias=False,
        )
        self.add_module(self.norm2_name, norm2)
        self.relu = nn.ReLU(inplace=True)

        # stage 1
        self.stage1_cfg = self.extra["stage1"]
        num_channels = self.stage1_cfg["num_channels"][0]
        block_type = self.stage1_cfg["block"]
        num_blocks = self.stage1_cfg["num_blocks"][0]

        block = self.blocks_dict[block_type]
        stage1_out_channels = num_channels * block.expansion
        self.layer1 = self._make_layer(block, stem_channels, num_channels, num_blocks)

        # stage 2
        self.stage2_cfg = self.extra["stage2"]
        num_channels = self.stage2_cfg["num_channels"]
        block_type = self.stage2_cfg["block"]

        block = self.blocks_dict[block_type]
        num_channels = [channel * block.expansion for channel in num_channels]
        self.transition1 = self._make_transition_layer(
            [stage1_out_channels],
            num_channels,
            strides[0],
        )
        self.stage2, pre_stage_channels = self._make_stage(
            self.stage2_cfg,
            num_channels,
            down_strides=strides,
        )

        # stage 3
        self.stage3_cfg = self.extra["stage3"]
        num_channels = self.stage3_cfg["num_channels"]
        block_type = self.stage3_cfg["block"]

        block = self.blocks_dict[block_type]
        num_channels = [channel * block.expansion for channel in num_channels]
        self.transition2 = self._make_transition_layer(
            pre_stage_channels,
            num_channels,
            strides[1],
        )
        self.stage3, pre_stage_channels = self._make_stage(
            self.stage3_cfg,
            num_channels,
            down_strides=strides,
        )

        # stage 4
        self.stage4_cfg = self.extra["stage4"]
        num_channels = self.stage4_cfg["num_channels"]
        block_type = self.stage4_cfg["block"]

        block = self.blocks_dict[block_type]
        num_channels = [channel * block.expansion for channel in num_channels]
        self.transition3 = self._make_transition_layer(
            pre_stage_channels,
            num_channels,
            strides[2],
        )
        self.stage4, pre_stage_channels = self._make_stage(
            self.stage4_cfg,
            num_channels,
            multiscale_output=multiscale_output,
            down_strides=strides,
        )

        self._freeze_stages()

    @property
    def norm1(self):
        """nn.Module: the normalization layer named "norm1" """
        return getattr(self, self.norm1_name)

    @property
    def norm2(self):
        """nn.Module: the normalization layer named "norm2" """
        return getattr(self, self.norm2_name)

    def _make_transition_layer(
        self,
        num_channels_pre_layer,
        num_channels_cur_layer,
        down_stride=2,
    ):
        """Make transition layer."""
        num_branches_cur = len(num_channels_cur_layer)
        num_branches_pre = len(num_channels_pre_layer)

        transition_layers = []
        for i in range(num_branches_cur):
            if i < num_branches_pre:
                if num_channels_cur_layer[i] != num_channels_pre_layer[i]:
                    transition_layers.append(
                        nn.Sequential(
                            build_conv_layer(
                                self.conv_cfg,
                                num_channels_pre_layer[i],
                                num_channels_cur_layer[i],
                                kernel_size=3,
                                stride=1,
                                padding=1,
                                bias=False,
                            ),
                            build_norm_layer(self.norm_cfg, num_channels_cur_layer[i])[
                                1
                            ],
                            nn.ReLU(inplace=True),
                        )
                    )
                else:
                    transition_layers.append(None)
            else:
                # Generates downsampled feature maps
                conv_downsamples = []
                for j in range(i + 1 - num_branches_pre):
                    in_channels = num_channels_pre_layer[-1]
                    out_channels = (
                        num_channels_cur_layer[i]
                        if j == i - num_branches_pre
                        else in_channels
                    )
                    conv_downsamples.append(
                        nn.Sequential(
                            build_conv_layer(
                                self.conv_cfg,
                                in_channels,
                                out_channels,
                                kernel_size=3,
                                stride=down_stride,
                                padding=1,
                                bias=False,
                            ),
                            build_norm_layer(self.norm_cfg, out_channels)[1],
                            nn.ReLU(inplace=True),
                        )
                    )
                transition_layers.append(nn.Sequential(*conv_downsamples))

        return nn.ModuleList(transition_layers)

    def _make_layer(self, block, inplanes, planes, blocks, stride=1):
        """Make each layer."""
        downsample = None
        if stride != 1 or inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                build_conv_layer(
                    self.conv_cfg,
                    inplanes,
                    planes * block.expansion,
                    kernel_size=1,
                    stride=stride,
                    bias=False,
                ),
                build_norm_layer(self.norm_cfg, planes * block.expansion)[1],
            )

        layers = []
        block_init_cfg = None
        if (
            self.pretrained is None
            and not hasattr(self, "init_cfg")
            and self.zero_init_residual
        ):
            if block is BasicBlock:
                block_init_cfg = dict(
                    type="Constant", val=0, override=dict(name="norm2")
                )
            elif block is Bottleneck:
                block_init_cfg = dict(
                    type="Constant", val=0, override=dict(name="norm3")
                )

        layers.append(
            block(
                inplanes,
                planes,
                stride,
                downsample=downsample,
                with_cp=self.with_cp,
                norm_cfg=self.norm_cfg,
                conv_cfg=self.conv_cfg,
                init_cfg=block_init_cfg,
            )
        )
        inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(
                block(
                    inplanes,
                    planes,
                    with_cp=self.with_cp,
                    norm_cfg=self.norm_cfg,
                    conv_cfg=self.conv_cfg,
                    init_cfg=block_init_cfg,
                )
            )

        return Sequential(*layers)

    def _make_stage(
        self,
        layer_config,
        in_channels,
        multiscale_output=True,
        down_strides=(2, 2, 2),
    ):
        """Make each stage."""
        num_modules = layer_config["num_modules"]
        num_branches = layer_config["num_branches"]
        num_blocks = layer_config["num_blocks"]
        num_channels = layer_config["num_channels"]
        block = self.blocks_dict[layer_config["block"]]

        hr_modules = []
        block_init_cfg = None
        if (
            self.pretrained is None
            and not hasattr(self, "init_cfg")
            and self.zero_init_residual
        ):
            if block is BasicBlock:
                block_init_cfg = dict(
                    type="Constant", val=0, override=dict(name="norm2")
                )
            elif block is Bottleneck:
                block_init_cfg = dict(
                    type="Constant", val=0, override=dict(name="norm3")
                )

        for i in range(num_modules):
            # multi_scale_output is only used for the last module
            if not multiscale_output and i == num_modules - 1:
                reset_multiscale_output = False
            else:
                reset_multiscale_output = True

            hr_modules.append(
                ModHRModule(
                    num_branches=num_branches,
                    blocks=block,
                    num_blocks=num_blocks,
                    in_channels=in_channels,
                    num_channels=num_channels,
                    multiscale_output=reset_multiscale_output,
                    down_strides=down_strides,
                    with_cp=self.with_cp,
                    norm_cfg=self.norm_cfg,
                    conv_cfg=self.conv_cfg,
                    block_init_cfg=block_init_cfg,
                )
            )

        return Sequential(*hr_modules), in_channels

    def _freeze_stages(self):
        """Freeze stages param and norm stats."""
        if self.frozen_stages >= 0:

            self.norm1.eval()
            self.norm2.eval()
            for m in [self.conv1, self.norm1, self.conv2, self.norm2]:
                for param in m.parameters():
                    param.requires_grad = False

        for i in range(1, self.frozen_stages + 1):
            if i == 1:
                m = getattr(self, f"layer{i}")
                t = getattr(self, f"transition{i}")
            elif i == 4:
                m = getattr(self, f"stage{i}")
            else:
                m = getattr(self, f"stage{i}")
                t = getattr(self, f"transition{i}")
            m.eval()
            for param in m.parameters():
                param.requires_grad = False
            t.eval()
            for param in t.parameters():
                param.requires_grad = False

    def _resize_and_concat(self, xs):
        xs = [
            resize(
                x,
                size=xs[0].shape[2:],
                mode=self.interp,
                align_corners=self.align_corners,
            )
            for x in xs
        ]
        x = torch.cat(xs, dim=1)
        return x

    def forward(self, x):
        """Forward function."""

        x = self.conv1(x)
        x = self.norm1(x)
        x = self.relu(x)
        x = self.conv2(x)
        x = self.norm2(x)
        x = self.relu(x)

        outs = []
        if self.return_stem:
            # add stem
            outs.append(x)

        # stage 1
        x = self.layer1(x)
        if self.return_sides:
            outs.append(x)

        # stage 2
        x_list = []
        for i in range(self.stage2_cfg["num_branches"]):
            if self.transition1[i] is not None:
                x_list.append(self.transition1[i](x))
            else:
                x_list.append(x)
        y_list = self.stage2(x_list)

        if self.return_sides:
            if self.resize_concat_sides[1]:
                outs.append(self._resize_and_concat(y_list))
            else:
                outs.append(y_list[0])

        # stage 3
        x_list = []
        for i in range(self.stage3_cfg["num_branches"]):
            if self.transition2[i] is not None:
                x_list.append(self.transition2[i](y_list[-1]))
            else:
                x_list.append(y_list[i])
        y_list = self.stage3(x_list)

        if self.return_sides:
            if self.resize_concat_sides[2]:
                outs.append(self._resize_and_concat(y_list))
            else:
                outs.append(y_list[0])

        # stage 4
        x_list = []
        for i in range(self.stage4_cfg["num_branches"]):
            if self.transition3[i] is not None:
                x_list.append(self.transition3[i](y_list[-1]))
            else:
                x_list.append(y_list[i])
        y_list = self.stage4(x_list)

        if self.return_sides:
            if self.resize_concat_sides[3]:
                outs.append(self._resize_and_concat(y_list))
            else:
                # might be redundant
                outs.append(y_list[0])

        outs += y_list
        return outs

    def train(self, mode=True):
        """Convert the model into training mode will keeping the normalization
        layer freezed."""
        super().train(mode)
        self._freeze_stages()
        if mode and self.norm_eval:
            for m in self.modules():
                # trick: eval have effect on BatchNorm only
                if isinstance(m, _BatchNorm):
                    m.eval()



# Copyright (c) ByteDance, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

from operator import xor
import os
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.utils.checkpoint as checkpoint

from visionsuite.cores.spatial_transform_decoupling.mmcv_custom import load_checkpoint
from mmdet.utils import get_root_logger
# from mmdet.models.builder import BACKBONES
from mmrotate.models.builder import ROTATED_BACKBONES
from visionsuite.cores.spatial_transform_decoupling.models import VisionTransformer as ViT


class PatchEmbed(nn.Module):
    """ Image to Patch Embedding
    """
    def __init__(self, img_size=[224, 224], patch_size=16, in_chans=3, embed_dim=768):
        super().__init__()
        self.num_patches_w = img_size[0] // patch_size
        self.num_patches_h = img_size[1] // patch_size

        num_patches = self.num_patches_w * self.num_patches_h
        self.img_size = img_size
        self.patch_size = patch_size
        self.num_patches = num_patches

        self.proj = nn.Conv2d(in_chans, embed_dim, kernel_size=patch_size, stride=patch_size)
            
    def forward(self, x, mask=None):
        B, C, H, W = x.shape
        return self.proj(x)


@ROTATED_BACKBONES.register_module()
class VisionTransformer(ViT):
    def __init__(self,
                 img_size,
                 patch_size,
                 embed_dim,
                 in_chans=3,
                 out_indices=[3, 5, 7, 11],
                 with_fpn=False,
                 with_simple_fpn=False,
                 frozen_stages=-1,
                 use_checkpoint=True,
                 learnable_pos_embed=True,
                 last_feat=False,
                 init_cfg=None,
                 **kwargs):
        super(VisionTransformer, self).__init__(
            img_size=img_size,
            patch_size=patch_size,
            in_chans=in_chans,
            embed_dim=embed_dim, 
            **kwargs)
        self.init_cfg = init_cfg

        assert not (with_fpn or with_simple_fpn) or (patch_size in (8, 16))
        assert not (with_fpn and with_simple_fpn)

        self.patch_size = patch_size
        self.last_feat = last_feat

        num_patches = self.patch_embed.num_patches
        self.pos_embed = nn.Parameter(
            torch.zeros(1, num_patches + 1, embed_dim),
            requires_grad=learnable_pos_embed,
        )
        
        self.with_fpn = with_fpn
        self.with_simple_fpn = with_simple_fpn
        self.frozen_stages = frozen_stages
        self.out_indices = out_indices
        self.use_checkpoint = use_checkpoint

        del self.norm, self.fc_norm, self.head
        if (with_fpn or with_simple_fpn) and patch_size == 16:
            self.fpn1 = nn.Sequential(
                nn.ConvTranspose2d(embed_dim, embed_dim, kernel_size=2, stride=2),
                nn.BatchNorm2d(embed_dim),
                nn.GELU(),
                nn.ConvTranspose2d(embed_dim, embed_dim, kernel_size=2, stride=2),
            )

            self.fpn2 = nn.Sequential(
                nn.ConvTranspose2d(embed_dim, embed_dim, kernel_size=2, stride=2),
            )

            self.fpn3 = nn.Identity()

            self.fpn4 = nn.MaxPool2d(kernel_size=2, stride=2)
        else:
            logger = get_root_logger()
            logger.info('Build model without FPN.')
        

    def train(self, mode=True):
        """Convert the model into training mode while keep layers freezed."""
        super(VisionTransformer, self).train(mode)
        self._freeze_stages()

    def _freeze_stages(self):
        if self.frozen_stages >= 0:
            self.patch_embed.eval()
            for param in self.patch_embed.parameters():
                param.requires_grad = False
            self.cls_token.requires_grad = False
            self.pos_embed.requires_grad = False
            self.pos_drop.eval()

        for i in range(1, self.frozen_stages + 1):
            if i  == len(self.blocks):
                norm_layer = getattr(self, 'norm') #f'norm{i-1}')
                norm_layer.eval()
                for param in norm_layer.parameters():
                    param.requires_grad = False

            m = self.blocks[i - 1]
            m.eval()
            for param in m.parameters():
                param.requires_grad = False

    #def init_weights(self, pretrained):
    #    logger = get_root_logger()
    #    if  os.path.isfile(pretrained):
    #        load_checkpoint(self, pretrained, strict=False, logger=logger)
    #    else:
    #        raise ValueError(f"checkpoint path {pretrained} is invalid")

    def init_weights(self):
        logger = get_root_logger()
        if self.init_cfg is None:
            raise ValueError(f'No pre-trained weights for {self.__class__.__name__}')
        else:
            assert 'checkpoint' in self.init_cfg, f'Only support ' \
                                                  f'specify `Pretrained` in ' \
                                                  f'`init_cfg` in ' \
                                                  f'{self.__class__.__name__} '
            pretrained = self.init_cfg['checkpoint']
            if  os.path.isfile(pretrained):
                load_checkpoint(self, pretrained, strict=False, logger=logger)
            else:
                raise ValueError(f"checkpoint path {pretrained} is invalid")

    def forward_encoder(self, x):
        for blk in self.blocks:
            x = blk(x)
        return x

    def forward(self, x):
        B, _, H, W = x.shape
        Hp, Wp = H // self.patch_size, W // self.patch_size
        x = self.prepare_tokens(x)
        features = []
        for i, blk in enumerate(self.blocks):
            if self.use_checkpoint:
                x = checkpoint.checkpoint(blk, x)
            else:
                x = blk(x)
            if i in self.out_indices:
                xp = x[:, 1:, :].permute(0, 2, 1).reshape(B, -1, Hp, Wp)       
                features.append(xp.contiguous())

        if self.with_fpn:
            ops = [self.fpn1, self.fpn2, self.fpn3, self.fpn4]
            for i in range(len(ops)):
                features[i] = ops[i](features[i])
        if self.with_simple_fpn:
            ops = [self.fpn1, self.fpn2, self.fpn3, self.fpn4]
            for i in range(len(ops)):
                features[i] = ops[i](features[-1])
        if self.last_feat:
            return tuple(features), x
        return tuple(features)



from visionsuite.engines.utils.registry.losses import TORCH_LOSSES
from visionsuite.engines.utils.registry.optims import TORCH_OPTIMIZERS
from visionsuite.engines.utils.registry.schedulers import TORCH_SCHEDULERS

from visionsuite.engines.utils.registry import RUNNERS as ROOT_RUNNERS
from visionsuite.engines.utils.registry import MODELS as ROOT_MODELS
from visionsuite.engines.utils.registry import DATASETS as ROOT_DATASETS
from visionsuite.engines.utils.registry import DATALOADERS as ROOT_DATALOADERS
from visionsuite.engines.utils.registry import LOOPS as ROOT_LOOPS
from visionsuite.engines.utils.registry import PIPELINES as ROOT_PIPELINES
from visionsuite.engines.utils.registry import SAMPLERS as ROOT_SAMPLERS
from visionsuite.engines.utils.registry import FUNCTIONALS as ROOT_FUNCTIONALS
from visionsuite.engines.utils.registry import TRAINERS as ROOT_TRAINERS
from visionsuite.engines.utils.registry import VALIDATORS as ROOT_VALIDATORS

from visionsuite.engines.utils.registry import Registry


scope = 'visionsuite.engines.classification.src'
RUNNERS = Registry('runners', parent=ROOT_RUNNERS, 
                   scope=scope,
                   locations=['visionsuite.engines.classification.src.runners']
        )

MODELS = Registry('models', parent=ROOT_MODELS, 
                   scope=scope,
                   locations=['visionsuite.engines.classification.src.models']
        )

OPTIMIZERS = Registry('optimizers', parent=TORCH_OPTIMIZERS, 
                   scope=scope,
                   locations=['visionsuite.engines.classification.src.optimizers']
        )

DATASETS = Registry('datasets', parent=ROOT_DATASETS, 
                   scope=scope,
                   locations=['visionsuite.engines.classification.src.datasets']
        )

DATALOADERS = Registry('dataloaders', parent=ROOT_DATALOADERS, 
                   scope=scope,
                   locations=['visionsuite.engines.classification.src.dataloaders']
        )

LOSSES = Registry('losses', parent=TORCH_LOSSES, 
                   scope=scope,
                   locations=['visionsuite.engines.classification.src.losses']
        )

SCHEDULERS = Registry('schedulers', parent=TORCH_SCHEDULERS, 
                   scope=scope,
                   locations=['visionsuite.engines.classification.src.schedulers']
        )

LOOPS = Registry('loops', parent=ROOT_LOOPS, 
                   scope=scope,
                   locations=['visionsuite.engines.classification.src.loops']
        )

PIPELINES = Registry('pipelines', parent=ROOT_PIPELINES, 
                   scope=scope,
                   locations=['visionsuite.engines.classification.src.pipelines']
        )

SAMPLERS = Registry('samplers', parent=ROOT_SAMPLERS, 
                   scope=scope,
                   locations=['visionsuite.engines.classification.src.samplers']
        )

FUNCTIONALS = Registry('functionals', parent=ROOT_FUNCTIONALS, 
                   scope=scope,
                   locations=['visionsuite.engines.classification.src.functionals']
        )

TRAINERS = Registry('trainers', parent=ROOT_TRAINERS, 
                   scope=scope,
                   locations=['visionsuite.engines.classification.src.trainers']
        )

VALIDATORS = Registry('validators', parent=ROOT_VALIDATORS, 
                   scope=scope,
                   locations=['visionsuite.engines.classification.src.validators']
        )



import matplotlib.pyplot as plt
import seaborn as sns


def autopct_with_counts(pct, allvals):
    absolute = int(pct/100.*sum(allvals))
    return "{:.1f}%\n({:d})".format(pct, absolute)

def draw_pie_chart(data, title, output_filename):
   sns.set_theme(style="whitegrid")
   plt.figure(figsize=(6, 6))
   plt.suptitle(title + f' ({sum(data.values())})')
   plt.pie(data.values(), labels=data.keys(), autopct=lambda pct: autopct_with_counts(pct, list(data.values())), startangle=140, colors=sns.color_palette("pastel"))
   plt.savefig(output_filename)
   
   
def draw_dist_chart(data, key, output_filename):
    plt.figure(figsize=(12, 10)) 
    plt.subplot(2, 1, 1)
    plt.hist(list(map(int, data['width'])), bins=range(min(list(map(int, data['width']))), max(list(map(int, data['width']))) + 2), edgecolor='black', alpha=0.7)
    # plt.xticks(range(min(list(map(int, data['width']))), max(list(map(int, data['width']))) + 1)) 
    plt.xlabel('width') 
    plt.ylabel('Frequency')
    plt.subplot(2, 1, 2)
    plt.hist(list(map(int, data['height'])), bins=range(min(list(map(int, data['height']))), max(list(map(int, data['height']))) + 2), edgecolor='black', alpha=0.7)
    # plt.xticks(range(min(list(map(int, data['height']))), max(list(map(int, data['height']))) + 1)) 
    plt.suptitle(f'Distribution of {key}') 
    plt.xlabel('height') 
    plt.ylabel('Frequency')
    plt.grid(True)  
    plt.savefig(output_filename)


"""This file only exists to be lazy-imported and avoid V2-related import warnings when just using V1."""
import torch
from torchvision import tv_tensors
from torchvision.transforms import v2


class PadIfSmaller(v2.Transform):
    def __init__(self, size, fill=0):
        super().__init__()
        self.size = size
        self.fill = v2._utils._setup_fill_arg(fill)

    def _get_params(self, sample):
        _, height, width = v2._utils.query_chw(sample)
        padding = [0, 0, max(self.size - width, 0), max(self.size - height, 0)]
        needs_padding = any(padding)
        return dict(padding=padding, needs_padding=needs_padding)

    def _transform(self, inpt, params):
        if not params["needs_padding"]:
            return inpt

        fill = v2._utils._get_fill(self.fill, type(inpt))
        fill = v2._utils._convert_fill_arg(fill)

        return v2.functional.pad(inpt, padding=params["padding"], fill=fill)


class CocoDetectionToVOCSegmentation(v2.Transform):
    """Turn samples from datasets.CocoDetection into the same format as VOCSegmentation.

    This is achieved in two steps:

    1. COCO differentiates between 91 categories while VOC only supports 21, including background for both. Fortunately,
       the COCO categories are a superset of the VOC ones and thus can be mapped. Instances of the 70 categories not
       present in VOC are dropped and replaced by background.
    2. COCO only offers detection masks, i.e. a (N, H, W) bool-ish tensor, where the truthy values in each individual
       mask denote the instance. However, a segmentation mask is a (H, W) integer tensor (typically torch.uint8), where
       the value of each pixel denotes the category it belongs to. The detection masks are merged into one segmentation
       mask while pixels that belong to multiple detection masks are marked as invalid.
    """

    COCO_TO_VOC_LABEL_MAP = dict(
        zip(
            [0, 5, 2, 16, 9, 44, 6, 3, 17, 62, 21, 67, 18, 19, 4, 1, 64, 20, 63, 7, 72],
            range(21),
        )
    )
    INVALID_VALUE = 255

    def _coco_detection_masks_to_voc_segmentation_mask(self, target):
        if "masks" not in target:
            return None

        instance_masks, instance_labels_coco = target["masks"], target["labels"]

        valid_labels_voc = [
            (idx, label_voc)
            for idx, label_coco in enumerate(instance_labels_coco.tolist())
            if (label_voc := self.COCO_TO_VOC_LABEL_MAP.get(label_coco)) is not None
        ]

        if not valid_labels_voc:
            return None

        valid_voc_category_idcs, instance_labels_voc = zip(*valid_labels_voc)

        instance_masks = instance_masks[list(valid_voc_category_idcs)].to(torch.uint8)
        instance_labels_voc = torch.tensor(instance_labels_voc, dtype=torch.uint8)

        # Calling `.max()` on the stacked detection masks works fine to separate background from foreground as long as
        # there is at most a single instance per pixel. Overlapping instances will be filtered out in the next step.
        segmentation_mask, _ = (instance_masks * instance_labels_voc.reshape(-1, 1, 1)).max(dim=0)
        segmentation_mask[instance_masks.sum(dim=0) > 1] = self.INVALID_VALUE

        return segmentation_mask

    def forward(self, image, target):
        segmentation_mask = self._coco_detection_masks_to_voc_segmentation_mask(target)
        if segmentation_mask is None:
            segmentation_mask = torch.zeros(v2.functional.get_size(image), dtype=torch.uint8)

        return image, tv_tensors.Mask(segmentation_mask)



#!/usr/bin/env python3

from .builder import DATASETS
from .custom import OTFCustomJointDataset

from pyEdgeEval.datasets.cityscapes_attributes import (
    CITYSCAPES_inst_labelIds,
    CITYSCAPES_labelIds,
    CITYSCAPES_label2trainId,
)

CITYSCAPES_CLASSES = (
    "road",
    "sidewalk",
    "building",
    "wall",
    "fence",
    "pole",
    "traffic light",
    "traffic sign",
    "vegetation",
    "terrain",
    "sky",
    "person",
    "rider",
    "car",
    "truck",
    "bus",
    "train",
    "motorcycle",
    "bicycle",
)
CITYSCAPES_PALETTE = [
    [128, 64, 128],
    [244, 35, 232],
    [70, 70, 70],
    [102, 102, 156],
    [190, 153, 153],
    [153, 153, 153],
    [250, 170, 30],
    [220, 220, 0],
    [107, 142, 35],
    [152, 251, 152],
    [70, 130, 180],
    [220, 20, 60],
    [255, 0, 0],
    [0, 0, 142],
    [0, 0, 70],
    [0, 60, 100],
    [0, 80, 100],
    [0, 0, 230],
    [119, 11, 32],
]
ignore_ids = [2, 3]


@DATASETS.register_module()
class OTFJointCityscapesDataset(OTFCustomJointDataset):
    """OTF edge generation for Cityscapes dataset

    - should only be used for training the model
    """

    CLASSES = CITYSCAPES_CLASSES
    PALETTE = CITYSCAPES_PALETTE

    def __init__(
        self,
        img_suffix="_leftImg8bit.png",
        seg_map_suffix="_gtFine_labelIds.png",
        inst_map_suffix="_gtFine_instanceIds.png",
        inst_sensitive=True,  # default instance sensitive
        labelIds=CITYSCAPES_labelIds,
        inst_labelIds=CITYSCAPES_inst_labelIds,
        ignore_indicies=[2, 3],
        label2trainId=CITYSCAPES_label2trainId,
        **kwargs,
    ):
        super(OTFJointCityscapesDataset, self).__init__(
            img_suffix=img_suffix,
            seg_map_suffix=seg_map_suffix,
            inst_map_suffix=inst_map_suffix,
            inst_sensitive=inst_sensitive,
            labelIds=labelIds,
            inst_labelIds=inst_labelIds,
            ignore_indices=ignore_indicies,
            label2trainId=label2trainId,
            **kwargs,
        )

    def format_results(self, **kwargs):
        raise ValueError("ERR: Should not use OTF for test set")

    def evaluate(self, **kwargs):
        raise ValueError("ERR: Should not use OTF for evaluation!")
